# Author: Christian Brodbeck <christianbrodbeck@nyu.edu>
from .ndvar import Butterworth, concatenate, convolve, correlation_coefficient, cross_correlation, cwt_morlet, dss, erode, filter_data, find_intervals, find_peaks, frequency_response, gaussian, label_operator, maximum, minimum, neighbor_correlation, normalize_in_cells, powerlaw_noise, psd_welch, rename_dim, resample, segment, set_adjacency, set_parc, set_time, set_tmin



"""Requires test data at root/test_data: gammatone.pickle, gammatone-left.pickle"""
import pytest

from eelbrain import load, gammatone_bank
from eelbrain.testing import assert_dataobj_equal
from eelbrain.testing.data import TEST_DATA_DIRECTORY, generate_sound


def test_gammatone_bank():
    target_path = TEST_DATA_DIRECTORY / 'gammatone.pickle'
    if not target_path.exists():
        pytest.skip(f"Test data missing: {target_path}")
    sound = generate_sound()

    target = load.unpickle(target_path)
    gt = gammatone_bank(sound, 20, 2000, 32, 1 / 100)
    assert_dataobj_equal(gt, target)

    target_left = load.unpickle(TEST_DATA_DIRECTORY / 'gammatone-left.pickle')
    gt = gammatone_bank(sound, 20, 2000, 32, 1 / 100, location='left')
    assert_dataobj_equal(gt, target_left)



# Author: Christian Brodbeck <christianbrodbeck@nyu.edu>
from eelbrain.pipeline import CombinationParc, SubParc


def test_combination_parc():
    parc = CombinationParc(
        'aparc',
        {'STG301': "split(transversetemporal + superiortemporal, 3)[:2]"}
    )
    assert parc._base_labels() == {'transversetemporal', 'superiortemporal'}
    parc = CombinationParc(
        'aparc',
        {
            'STG301': "split(transversetemporal + superiortemporal, 3)[:2]",
            'MTG301': "split(middletemporal, 3)[:2]",
        }
    )
    assert parc._base_labels() == {'transversetemporal', 'superiortemporal', 'middletemporal'}


def test_sub_parc():
    parc = SubParc('aparc', ('transversetemporal', 'superiortemporal'))
    assert parc._base_labels() == {'transversetemporal', 'superiortemporal'}



# skip test: data unavailable
from eelbrain.pipeline import *


class Somatomotor(MneExperiment):

    ignore_entities = {
        'ignore_runs': ('2', '3'),
        'ignore_sessions': 'mri',
    }

    raw = {
        'tsss': RawMaxwell('raw', st_duration=10., ignore_ref=True, st_correlation=0.9, st_only=True),
        '1-40': RawFilter('tsss', 1, 40),
        'ica': RawICA('1-40', 'somatomotor', n_components=0.99),
    }

    variables = {
        'is_null': LabelVar('trigger', {2: 'true', (1, 3): 'false'}),
        'stimulus': LabelVar('trigger', {2: 'null', 1: 'finger', 3: 'somatosensory'}),
    }

    epochs = {
        'not_null': PrimaryEpoch('somatomotor', "is_null == 'false'", samplingrate=251.005),
        'finger': SecondaryEpoch('not_null', "stimulus == 'finger'"),
        'somatosensory': SecondaryEpoch('not_null', "stimulus == 'somatosensory'"),
    }

    tests = {
        '=0': TTestOneSample(),
        'connection': TTestRelated('stimulus', 'finger', 'somatosensory'),
        'anova': ANOVA('stimulus * subject'),
    }


root = '/mnt/d/Data/Somatomotor'
e = Somatomotor(root)
e.set(rej='')
# print(e.load_evoked(subjects=-1, data='meg'))
print(e.load_test('connection', 0.3, 0.5, 0.05, data='meg', baseline=False, epoch='not_null', make=True))



# Author: Christian Brodbeck <christianbrodbeck@nyu.edu>
"""Parse contrast expressions

AST
===

Nested tuple composed of:
- comparison:  ``('comp', cell_1, cell_0)``
- Unary functions:  ``('ufunc', func, arg)``
- Binary functions:  ``('bfunc', func, [arg1, arg2])``
- Array functions:  ``('afunc', func, [arg1, arg2, ...])``
where ``arg1`` etc. are in turn comparisons and functions.


Grammar
=======

Terminals
---------

- name
- comparator:  <, >
- operator:  +, -
- , ( ) |


Rules
-----

- SPM -> cell comparator cell
- SPM -> ufunc ( SPM )
- SPM -> bfunc ( SPM , SPM )
- SPM -> nfunc ( args )
- SPM -> SPM operator SPM

- args -> SPM
- args -> SPM , args

- cell -> name
- cell -> name | cell


"""
import re
# from token import LPAR, RPAR, NAME, COMMA, GREATER, LESS, VBAR

import numpy as np

SYNTAX_CHARS = '()|<>,+-'

# array functions:  work on array, take axis argument
np_afuncs = {'min': np.min,
             'max': np.max,
             'sum': np.sum}
# binary functions:  work on two arrays
np_bfuncs = {'subtract': np.subtract,
             'add': np.add}
# unary functions:  work on a single array
np_ufuncs = {'abs': np.abs,
             'negative': np.negative}

FUNCS = {name: (kind, func) for funcs, kind in
         ((np_afuncs, 'afunc'), (np_bfuncs, 'bfunc'), (np_ufuncs, 'ufunc'))
         for name, func in funcs.items()}
OP_FUNCS = {'-': np.subtract, '+': np.add}


RE_NAME = re.compile(r"[*\w\d ]*[*\w\d]")


class Streamer:

    def __init__(self, expression):
        tokens, indexes = tokenize(expression)
        assert len(indexes) == len(tokens)
        self._expression = expression
        self._tokens = tokens
        self._indexes = indexes
        self._i = -1
        self._imax = len(tokens) - 1
        self.eof = False

    @property
    def current(self):
        return self._tokens[self._i]

    def __iter__(self):
        return self

    def __next__(self, increment=1):
        self._i += increment
        if self._i >= self._imax:
            if self._i == self._imax:
                self.eof = True
            else:
                raise StopIteration
        return self._tokens[self._i]

    def lookahead(self, increment=1):
        i = self._i + increment
        if i > self._imax:
            return None
        else:
            return self._tokens[i]

    def error(self, desc):
        raise SyntaxError(desc, (None, 0, self._indexes[self._i] + 1,
                                 self._expression))


def tokenize(expression):
    i = 0
    imax = len(expression)
    tokens = []
    indexes = []
    while i < imax:
        if expression[i].isspace():
            i += 1
            continue
        elif expression[i] in SYNTAX_CHARS:
            tokens.append(expression[i])
            indexes.append(i)
            i += 1
            continue
        m = RE_NAME.match(expression, i)
        if m:
            tokens.append(m.group())
            indexes.append(i)
            i = m.end()
            continue
        raise SyntaxError("Invalid contrast expression syntax",
                          (None, 0, i + 1, expression))
    return tokens, indexes


def parse_spm(streamer, process_op=True):
    # parse function
    lookahead = streamer.lookahead()
    if lookahead in FUNCS and streamer.lookahead(2) == '(':
        kind, func = FUNCS[next(streamer)]
        next(streamer)
        args = parse_args(streamer)
        if kind == 'ufunc':
            if len(args) != 1:
                streamer.error("%s function needs exactly 1 argument" %
                               func.__name__)
            args = args[0]
        elif kind == 'bfunc':
            if len(args) != 2:
                streamer.error("%s function needs exactly 2 arguments" %
                               func.__name__)
        elif len(args) < 2:
            streamer.error("%s function needs at least 2 arguments" %
                           func.__name__)
        return kind, func, args
    elif lookahead == '(':
        next(streamer)
        out = parse_spm(streamer)
        if next(streamer) != ')':
            streamer.error("Expected )")
    else:
        # beginning of cell comparator cell
        cell_1 = parse_cell(streamer)
        comp = next(streamer)
        if comp not in '<>':
            streamer.error("Expected '<' or '>'")
        cell_2 = parse_cell(streamer)
        if comp == '<':
            cell_1, cell_2 = cell_2, cell_1
        out = ('comp', cell_1, cell_2)

    if not process_op:
        return out
    # look for operator
    while (not streamer.eof) and streamer.lookahead() not in ',)':
        op = next(streamer)
        if op in OP_FUNCS:
            arg = parse_spm(streamer, False)
            out = ('bfunc', OP_FUNCS[op], (out, arg))
        else:
            streamer.error("Expected ), + or -")
    return out


def parse_cell(streamer):
    cell = [next(streamer)]
    while streamer.lookahead() == '|':
        next(streamer)
        cell.append(next(streamer))
    if len(cell) == 1:
        return cell[0]
    else:
        return tuple(cell)


def parse_args(streamer):
    args = [parse_spm(streamer)]
    delim = next(streamer)
    while delim == ',':
        args.append(parse_spm(streamer))
        delim = next(streamer)
    if delim != ')':
        streamer.error("Expected ')'")
    return tuple(args)


def parse(expression):
    streamer = Streamer(expression)
    return parse_spm(streamer)



# make mne backwards-compatibility
import packaging.version

import mne


MNE_VERSION = packaging.version.parse(mne.__version__)
V0_19 = packaging.version.parse('0.19')
V0_22 = packaging.version.parse('0.22')
V0_24 = packaging.version.parse('0.24')
V1 = packaging.version.parse('1')



"""Exceptions used throughout Eelbrain"""
from typing import Collection

from ._text import enumeration, plural


class KeysMissing(KeyError):
    "A styling dictionary is missing a key (more information than KeyError)"
    def __init__(self, keys: Collection, from_name: str, from_dict: dict):
        KeyError.__init__(self, keys, from_name, from_dict)

    def __str__(self):
        keys, from_name, from_dict = self.args
        n = len(keys)
        return f"{plural('Key', n)} {enumeration(map(repr, keys))} missing from {from_name}={from_dict!r}"


class DefinitionError(Exception):
    "MneExperiment definition error"


class EvalError(Exception):
    "Error while evaluating expression"
    def __init__(self, expression, exception, context):
        Exception.__init__(self, f"Error evaluating {expression!r} in {context}: {exception}")


class DimensionMismatchError(Exception):
    "Trying to align NDVars with mismatching dimensions"

    @classmethod
    def from_dims_list(cls, message, dims_list, check: bool):
        from ._data_obj import dims_stackable

        unique_dims = []
        for dims in dims_list:
            if any(dims_stackable(dims, dims_, check) for dims_ in unique_dims):
                continue
            else:
                unique_dims.append(dims)
        desc = '\n'.join(map(str, unique_dims))
        return cls(f'{message}\n{desc}')


class WrongDimensionError(Exception):
    "Dimension that is supported"


class IncompleteModelError(Exception):
    "Function requires a fully specified model"


class OldVersionError(Exception):
    "Trying to load a file from a version that is no longer supported"


class ZeroVarianceError(Exception):
    "Data with zero variance"



# Author: Christian Brodbeck <christianbrodbeck@nyu.edu>
# cython: boundscheck=False, wraparound=False, cdivision=True, initializedcheck=False, language_level=3
cimport numpy


ctypedef numpy.float64_t FLOAT64
ctypedef numpy.int64_t INT64

cdef extern from "math.h":
    double round(double x)


def aggregate_left(
        FLOAT64 [:] xf,
        int n_samples,
        float step,
        int n_window,
        FLOAT64 [:] out,
):
    cdef:
        Py_ssize_t i, j, start
        Py_ssize_t n_xf = xf.shape[0]

    for i in range(n_samples):
        start = <Py_ssize_t>round(i * step)
        stop = start + n_window
        if stop > n_xf:
            stop = n_xf
        for j in range(start, stop):
            out[i] += xf[j]


def aggregate_right(
        FLOAT64 [:] xf,
        int n_samples,
        float step,
        int n_window,
        FLOAT64 [:] out,
):
    cdef:
        Py_ssize_t i, j, stop

    for i in range(n_samples):
        stop = <Py_ssize_t>round(i * step) + 1
        start = stop - n_window
        if start < 0:
            start = 0
        for j in range(start, stop):
            out[i] += xf[j]



# Author: Christian Brodbeck <christianbrodbeck@nyu.edu>
from dataclasses import replace
from itertools import product
from math import floor
from pathlib import Path
from warnings import catch_warnings, filterwarnings

import numpy as np
from numpy.testing import assert_array_equal, assert_allclose
import pickle
import pytest
from pytest import approx
import scipy.io
import scipy.stats
from eelbrain import datasets, boosting, combine, convolve, correlation_coefficient, epoch_impulse_predictor, NDVar, UTS, Scalar

from eelbrain.testing import assert_dataobj_equal
from eelbrain._trf._boosting import Boosting, DeconvolutionData, Split, convolve_1d
from eelbrain._trf._boosting_opt import boosting_fit


DATA_DIR = Path(__file__).parent


def assert_res_equal(res1, res):
    assert_array_equal(res1.h, res.h)
    assert res1.r == res.r
    assert res1.r_rank == res.r_rank


def convolve_array(
        h: np.ndarray,  # (n_stims, h_n_samples)
        x: np.ndarray,  # (n_stims, n_samples)
        x_pads: np.ndarray,  # (n_stims,)
        h_i_start: int = 0,
) -> np.ndarray:
    n_x, n_times = x.shape
    segments = np.array(((0, n_times),), np.int64)
    out = np.empty(n_times)
    convolve_1d(h, x, x_pads, h_i_start, segments, out)
    return out


def test_boosting():
    "Test boosting NDVars"
    ds = datasets._get_continuous(ynd=True)
    y = ds['y']
    ynd = ds['ynd']
    x1 = ds['x1']
    x2 = ds['x2']
    y_mean = y.mean()
    x2_mean = x2.mean('time')

    # test values from running function, not verified independently
    res = boosting(y, x1 * 2000, 0, 1, scale_data=False, mindelta=0.0025, debug=True)
    assert repr(res) == '<BoostingResult y ~ x1, 0 - 1, scale_data=False, mindelta=0.0025>'
    assert_allclose(res.h.x, [0, 0, 0, 0.0025, 0, 0, 0, 0, 0, 0.001875], atol=1e-6)
    assert res.r == approx(0.75, abs=0.001)
    assert res.y_mean is None
    assert res.h.info['unit'] == 'V'
    assert res.h_scaled.info['unit'] == 'V'
    assert res.residual == approx(((y[.9:] - res.y_pred[.9:])**2).sum())
    with pytest.raises(NotImplementedError):
        _ = res.proportion_explained

    res = boosting(y, x1, 0, 1)
    assert repr(res) == '<BoostingResult y ~ x1, 0 - 1>'
    assert res.r == approx(0.83, abs=0.001)
    assert res.y_mean == y_mean
    assert res.y_scale == y.std()
    assert res.x_mean == x1.mean()
    assert res.x_scale == x1.std()
    assert res.h.name == 'x1'
    assert res.h.info['unit'] == 'normalized'
    assert res.h_scaled.name == 'x1'
    assert res.h_scaled.info['unit'] == 'V'
    assert res.proportion_explained == approx(0.506, abs=0.001)
    # inplace
    res_ip = boosting(y.copy(), x1.copy(), 0, 1, 'inplace')
    assert_res_equal(res_ip, res)
    # persistence
    res_p = pickle.loads(pickle.dumps(res, pickle.HIGHEST_PROTOCOL))
    assert_res_equal(res_p, res)

    # L1 error
    res = boosting(y, x1 * 2000, 0, 1, error='l1', debug=True)
    assert res.residual == ((y[.9:] - res.y_mean) / res.y_scale - res.y_pred[.9:]).abs().sum()
    res_ndb = boosting(y, x1 * 2000, 0, 1, error='l1')
    assert res_ndb.residual == res.residual

    # cross-validation
    res = boosting(y[:7.5], x1[:7.5], 0, 1, scale_data=False, partitions=3, debug=True)
    res_cv = boosting(y, x1, 0, 1, test=1, scale_data=False, partitions=4, debug=True, partition_results=True)
    assert correlation_coefficient(res.h, res_cv.h) == approx(.986, abs=.001)
    # using cross-prediction
    y_pred = res_cv.cross_predict(x1, scale='normalized')
    assert correlation_coefficient(y_pred, y, 'time') == pytest.approx(res_cv.r)
    # fit res_cv on same data as res
    fit = res_cv.fit
    fit.data.splits = replace(fit.data.splits, splits=[fit.data.splits.splits[i] for i in [9, 10, 11]])
    fit.fit(0, 1, 0, 'l2')
    res_cv = fit.evaluate_fit(debug=True)
    assert_dataobj_equal(res_cv.h, res.h)
    y_pred = convolve(res_cv.h_scaled, x1[7.5:])
    assert correlation_coefficient(y_pred, y[7.5:]) == approx(res_cv.r, abs=1e-8)

    res = boosting(y, x2, 0, 1)
    assert res.r == approx(0.601, abs=0.001)
    assert res.proportion_explained == approx(0.273, abs=0.001)

    res = boosting(y, x2, 0, 1, error='l1')
    assert res.r == approx(0.553, abs=0.001)
    assert res.y_mean == y.mean()
    assert res.y_scale == (y - y_mean).abs().mean()
    assert_array_equal(res.x_mean.x, x2_mean)
    assert_array_equal(res.x_scale, (x2 - x2_mean).abs().mean('time'))
    assert res.proportion_explained == approx(0.123, abs=0.001)

    # 2 predictors
    res = boosting(y, [x1, x2], 0, 1)
    assert res.r == approx(0.947, abs=0.001)
    # selective stopping
    res = boosting(y, [x1, x2], 0, 1, selective_stopping=1)
    assert res.r == approx(0.967, abs=0.001)
    res = boosting(y, [x1, x2], 0, 1, selective_stopping=2)
    assert res.r == approx(0.992, abs=0.001)

    # 2d-y
    res_cv = boosting('ynd', ['x1', 'x2'], 0, 1, data=ds, test=1, partitions=4, partition_results=True)
    y_pred = res_cv.cross_predict(data=ds)
    assert_dataobj_equal(correlation_coefficient(y_pred, ynd, 'time'), res_cv.r, 2, name=False)


@pytest.mark.parametrize('error', ['l1', 'l2'])
def test_boosting_cross_predict(error):
    """Test cross-predicting data using stored TRFs"""
    ds = datasets._get_continuous()

    # Without scaling
    trf = boosting('y', 'x1', 0, 1, data=ds, error=error, partitions=3, test=1, partition_results=True, debug=True, scale_data=False)
    y_pred = trf.cross_predict('x1', ds)
    assert_array_equal(y_pred, trf.y_pred)

    # With scaling: normalized
    trf = boosting('y', 'x1', 0, 1, data=ds, error=error, partitions=3, test=1, partition_results=True, debug=True)
    # Results
    assert getattr(trf, f'{error}_total') == pytest.approx(trf.n_samples, 1e-16)
    # Re-predict
    y_pred = trf.cross_predict('x1', ds, scale='normalized')
    assert_array_equal(y_pred, trf.y_pred)
    # Proportion explained
    y_normalized = ds['y'] - trf.y_mean
    y_normalized /= trf.y_scale
    y_residual = y_normalized - y_pred
    if error == 'l1':
        proportion_explained = 1 - (y_residual.abs().sum('time') / y_normalized.abs().sum('time'))
    else:
        proportion_explained = 1 - ((y_residual ** 2).sum('time') / (y_normalized ** 2).sum('time'))
    assert proportion_explained == pytest.approx(trf.proportion_explained, 1e-16)

    # With scaling: original scale
    y_pred = trf.cross_predict('x1', ds)
    assert_array_equal(y_pred, trf.y_pred * trf.y_scale + trf.y_mean)


def test_boosting_epochs():
    """Test boosting with epoched data"""
    ds = datasets.get_uts(True, vector3d=True)
    p1 = epoch_impulse_predictor('uts', 'A=="a1"', name='a1', data=ds)
    p0 = epoch_impulse_predictor('uts', 'A=="a0"', name='a0', data=ds)
    p1 = p1.smooth('time', .05, 'hamming')
    p0 = p0.smooth('time', .05, 'hamming')
    # 1d
    for tstart, basis in product((-0.1, 0.1, 0), (0, 0.05)):
        print(f"{tstart=}, {basis=}")
        res = boosting('uts', [p0, p1], tstart, 0.6, model='A', data=ds, basis=basis, partitions=3, debug=True)
        assert res.r == approx(0.238, abs=2e-3)
        y = convolve(res.h_scaled, [p0, p1], name='predicted')
        assert correlation_coefficient(y, res.y_pred) > .999
        assert y.name == 'predicted'
        r = correlation_coefficient(y, ds['uts'])
        assert res.r == approx(r, abs=1e-3)
        assert res.splits.n_partitions == 3
    # 2d
    res = boosting('utsnd', [p0, p1], 0, 0.6, model='A', data=ds, partitions=3)
    assert len(res.h) == 2
    assert res.h[0].shape == (5, 60)
    assert res.h[1].shape == (5, 60)
    y = convolve(res.h_scaled, [p0, p1])
    r = correlation_coefficient(y, ds['utsnd'], ('case', 'time'))
    assert_dataobj_equal(res.r, r, decimal=3, name=False)
    # cross-validation
    res_cv = boosting('utsnd', [p0, p1], 0, 0.6, error='l1', data=ds, partitions=3, test=1, partition_results=True, debug=True)
    y_pred = res_cv.cross_predict([p0, p1], scale='normalized')
    assert_array_equal(y_pred, res_cv.y_pred)
    # vector
    res = boosting('v3d', [p0, p1], 0, 0.6, error='l1', model='A', data=ds, partitions=3)
    assert res.residual.ndim == 0


def test_boosting_object():
    "Test Boosting object"
    ds = datasets._get_continuous(ynd=True)
    data = DeconvolutionData('y', 'x2', ds)
    data.apply_basis(0.2, 'hamming')
    data.normalize('l1')
    data.initialize_cross_validation(4, test=1)
    model = Boosting(data)
    model.fit(0, 1, selective_stopping=1, error='l1')
    res_oo = model.evaluate_fit()

    res_func = boosting('y', 'x2', 0, 1, data=ds, error='l1', basis=0.2, partitions=4, test=1, selective_stopping=1)
    assert res_oo.r == res_func.r

    res_part = model.evaluate_fit(partition_results=True)
    assert len(res_part.partition_results) == 4
    # h with basis
    h_part = combine([res.h for res in res_part.partition_results])
    assert len(h_part) == 4
    assert_dataobj_equal(h_part.mean('case'), res_oo.h, decimal=10)
    # raw h
    h_part = combine([res._h for res in res_part.partition_results])
    assert_dataobj_equal(h_part.mean('case'), res_oo._h)


def test_result():
    "Test boosting results"
    ds = datasets._get_continuous()
    x1 = ds['x1']

    # convolve function
    y = convolve([ds['h1'], ds['h2']], [ds['x1'], ds['x2']])
    assert_dataobj_equal(y, ds['y'], name=False)

    # test prediction with res.h and res.h_scaled
    res = boosting(ds['y'], ds['x1'], 0, 1)
    y1 = convolve(res.h_scaled, ds['x1'])
    x_scaled = ds['x1'] / res.x_scale
    y2 = convolve(res.h, x_scaled)
    y2 *= res.y_scale
    y2 += y1.mean() - y2.mean()  # mean can't be reconstructed
    assert_dataobj_equal(y1, y2, decimal=12)
    # reconstruction
    res = boosting(x1, y, -1, 0, debug=True)
    x1r = convolve(res.h_scaled, y)
    assert correlation_coefficient(res.y_pred, x1r) > .999
    assert correlation_coefficient(x1r[0.9:], x1[0.9:]) == approx(res.r, abs=1e-3)

    # test NaN checks  (modifies data)
    ds['x2'].x[1, 50] = np.nan
    with pytest.raises(ValueError):
        boosting(ds['y'], ds['x2'], 0, .5)
    with pytest.raises(ValueError):
        boosting(ds['y'], ds['x2'], 0, .5, False)
    ds['x2'].x[1, :] = 1
    with catch_warnings():
        filterwarnings('ignore', category=RuntimeWarning)
        with pytest.raises(ValueError):
            boosting(ds['y'], ds['x2'], 0, .5)
        ds['y'].x[50] = np.nan
        with pytest.raises(ValueError):
            boosting(ds['y'], ds['x1'], 0, .5)
        with pytest.raises(ValueError):
            boosting(ds['y'], ds['x1'], 0, .5, False)


def evaluate_kernel(y, y_pred, test_seg_len, n_skip):
    y = y[n_skip: test_seg_len]
    y_pred = y_pred[n_skip:]
    return np.corrcoef(y, y_pred)[0, 1], scipy.stats.spearmanr(y, y_pred)[0]


def test_boosting_fit():
    "Test boosting() against svdboostV4pred.m"
    # 1d-TRF
    mat = scipy.io.loadmat(DATA_DIR / 'test_boosting.mat')
    y = mat['signal'][0]
    x = mat['stim']
    x_pads = np.zeros(len(x))

    y_len = len(y)
    seg_len = int(y_len / 40)
    all_segments = np.array([[0, seg_len], [seg_len, y_len]], np.int64)
    split = Split(all_segments[1:], all_segments[:1])
    tstart = np.array([0], np.int64)
    tstop = np.array([10], np.int64)
    h, history = boosting_fit(y, x, x_pads, split.train, split.validate, split.train_and_validate, tstart, tstop, 0.005, 0.005, 2)
    test_sse_history = [step.e_test for step in history]
    test_seg_len = int(floor(x.shape[1] / 40))
    y_pred = convolve_array(h, x[:, :test_seg_len], x_pads)
    r, rr = evaluate_kernel(y, y_pred, test_seg_len, h.shape[1] - 1)

    assert_allclose(test_sse_history, mat['Str_testE'][0])
    assert_array_equal(h, mat['h'])
    assert r == approx(mat['crlt'][0, 0])
    assert rr == approx(mat['crlt'][1, 0])

    # 2d-TRF
    mat = scipy.io.loadmat(DATA_DIR / 'test_boosting_2d.mat')
    y = mat['signal'][0]
    x = mat['stim']
    x_pads = np.zeros(len(x))
    tstart = np.array([0, 0, 0], np.int64)
    tstop = np.array([10, 10, 10], np.int64)
    h, history = boosting_fit(y, x, x_pads, split.train, split.validate, split.train_and_validate, tstart, tstop, 0.005, 0.005, 2)
    test_sse_history = [step.e_test for step in history]
    test_seg_len = int(floor(x.shape[1] / 40))
    y_pred = convolve_array(h, x[:, :test_seg_len], x_pads)
    r, rr = evaluate_kernel(y, y_pred, test_seg_len, h.shape[1] - 1)

    # svdboostV4pred multiplies error by number of predictors
    assert_allclose(test_sse_history, mat['Str_testE'][0] / 3)
    assert_array_equal(h, mat['h'])
    assert r == approx(mat['crlt'][0, 0])
    assert rr == approx(mat['crlt'][1, 0])


def test_trf_len():
    # test vanilla boosting
    rng = np.random.RandomState(0)
    x = NDVar(rng.normal(0, 1, 1000), UTS(0, 0.1, 1000), name='x')
    k = NDVar(rng.randint(0, 10, 5) / 10, UTS(0, 0.1, 5), name='k')
    y = convolve(k, x, name='y')
    res = boosting(y, x, 0, 0.5, partitions=3)
    assert correlation_coefficient(res.h, k) > 0.99
    assert repr(res) == '<BoostingResult y ~ x, 0 - 0.5, partitions=3>'
    # split the predictor into two complementary time windows (should be identical)
    res2 = boosting(y, [x, x], [0.000, 0.250], [0.250, 0.500], partitions=3)
    assert_array_equal(res2.h[0] + res2.h[1], res.h)
    assert res2.r == res.r

    # test multiple tstart, tend
    x2 = NDVar(rng.normal(0, 1, 1000), UTS(0, 0.1, 1000), name='x2')
    k2 = NDVar(rng.randint(0, 10, 4) / 10, UTS(-0.1, 0.1, 4), name='k2')
    y2 = y + convolve(k2, x2)
    res = boosting(y2, [x, x2], [0, -0.1], [0.5, 0.3], partitions=3)
    assert correlation_coefficient(res.h[0].sub(time=(0, 0.5)), k) > 0.99
    assert correlation_coefficient(res.h[1].sub(time=(-0.1, 0.3)), k2) > 0.99
    assert repr(res) == '<BoostingResult y ~ x (0 - 0.5) + x2 (-0.1 - 0.3), partitions=3>'

    # test scalar res.tstart res.tstop
    res = boosting(y2, [x, x2], 0, 0.5, partitions=3)
    assert res.tstart == 0
    assert res.tstop == 0.5

    # test duplicate tstart, tend for multiple predictors
    k3 = NDVar(k2.x, UTS(0, 0.1, 4))
    y3 = convolve(k, x) + convolve(k3, x2)
    res = boosting(y3, [x, x2], 0, 0.5, partitions=3)
    assert correlation_coefficient(res.h[0], k) > 0.99
    assert correlation_coefficient(res.h[1].sub(time=(0, 0.4)), k3) > 0.99

    # test vanilla boosting with 2d predictor
    x4 = NDVar(rng.normal(0, 1, (2, 1000)), (Scalar('xdim', [1, 2]), UTS(0, 0.1, 1000)))
    k4 = NDVar(rng.randint(0, 10, (2, 5)) / 10, (Scalar('xdim', [1, 2]), UTS(0, 0.1, 5)))
    y4 = convolve(k4, x4)
    res = boosting(y4, x4, 0, 0.5, partitions=3)
    assert correlation_coefficient(res.h, k4) > 0.99

    # test multiple tstart, tstop with 1d, 2d predictors
    y5 = y4 + y2
    res = boosting(y5, [x, x2, x4], [0, -0.1, 0], [0.5, 0.3, 0.5], partitions=3)
    assert correlation_coefficient(res.h[0].sub(time=(0, 0.5)), k) > 0.99
    assert correlation_coefficient(res.h[1].sub(time=(-0.1, 0.3)), k2) > 0.99
    assert correlation_coefficient(res.h[2].sub(time=(0, 0.5)), k4) > 0.99

    # tests tstart/tstop for each time series (not implemented)
    # res2 = boosting(y5, [x, x2, x4], [0, -0.1, 0, 0], [0.5, 0.3, 0.5, 0.5])
    # assert_array_equal(res.h[0], res2.h[0])
    # assert_array_equal(res.h[1], res2.h[1])
    # assert_array_equal(res.h[2], res2.h[2])



# Author: Christian Brodbeck <christianbrodbeck@nyu.edu>
import numpy as np
import pytest

from eelbrain._trf._fit_metrics import error_for_indexes


# numpy-based error functions
#############################
def np_l2(x):
    return np.dot(x, x[:, None])[0]


def np_l1(x):
    return np.abs(x).sum()


# test function
###############
def test_error_functions():
    "Test optimized error functions"
    x = np.random.normal(0., 1., 100)
    index = np.array(((0, 100),), np.int64)
    assert error_for_indexes(x, index, 1) == pytest.approx(np_l1(x))
    assert error_for_indexes(x, index, 2) == pytest.approx(np_l2(x))



# ---
# jupyter:
#   jupytext:
#     text_representation:
#       extension: .py
#       format_name: sphinx
#       format_version: '1.1'
#       jupytext_version: 1.10.2
#   kernelspec:
#     display_name: Python 3
#     language: python
#     name: python3
# ---

"""
.. _exa-customizing-plots:

Customizing plots
=================

With the exception of :mod:`plot.brain` plots, Eelbrain's plots are all based 
on :mod:`matplotlib`. A lot of fine control over the plots can be achieved 
through two means:

 - Customizing Matplotlib globally, before calling Eelbrain plotting functions, 
   through `styles or ``rcParams`` <https://matplotlib.org/tutorials/
   introductory/customizing.html#customizing-matplotlib-with-style-sheets-and-rcparams>`_
 - Accessing and modifying components of the plots after calling Eelbrain plotting 
   functions
"""
# sphinx_gallery_thumbnail_number = 3
from eelbrain import *
import matplotlib.style

ds = datasets.get_uv()

###############################################################################
# Styles
# ^^^^^^
# Matplotlib offers several styles

p = plot.Boxplot('fltvar', 'A % B', match='rm', data=ds, w=2)

# Apply a style
matplotlib.style.use('ggplot')
p = plot.Boxplot('fltvar', 'A % B', match='rm', data=ds, w=2)

matplotlib.style.use('bmh')
p = plot.Boxplot('fltvar', 'A % B', match='rm', data=ds, w=2)

###############################################################################
# ``rcParams``
# ^^^^^^^^^^^^
# Individual styles parameters can be modified directly in``rcParams``

matplotlib.rcParams['font.family'] = 'serif'
matplotlib.rcParams['font.size'] = 8
p = plot.Boxplot('fltvar', 'A % B', match='rm', data=ds, w=2)

# revert back to the default style
matplotlib.style.use('default')

###############################################################################
# Modifying components
# ^^^^^^^^^^^^^^^^^^^^
# Matplotlib can be used to fully customize a plot's appearance by accessing
# the underlying :class:`matplotlib.figure.Figure` object through the plot's
# :attr:`.figure` aatribute.

p = plot.Boxplot('fltvar', 'A % B', match='rm', data=ds, w=2)

p = plot.Boxplot('fltvar', 'A % B', match='rm', data=ds, w=2, h=3, xlabel=False)
ax = p.figure.axes[0]
ax.set_xticklabels(['A long label', 'An even longer label', 'Another label', 'And yet another one'], rotation=45, ha='right')
ax.grid(axis='y')
ax.set_yticks([-2, 0, 2])
ax.tick_params('y', left=False)
for spine in ax.spines.values():
    spine.set_visible(False)
p.figure.tight_layout()



# Author: Christian Brodbeck <christianbrodbeck@nyu.edu>
"""Plot uniform time-series of one variable."""
from typing import Any, Callable, Dict, Sequence, Tuple, Union

import matplotlib.axes
import matplotlib.colors
import matplotlib.patches
import numpy as np

from .._colorspaces import oneway_colors
from .._data_obj import NDVarArg, Case, CategorialArg, CellArg, IndexArg, Dataset, NDVar, asndvar, cellname, longname
from .._stats.testnd import NDTest
from .._utils import deprecate_ds_arg
from . import _base
from ._base import AxisData, EelFigure, Layout, LegendArg, LegendMixin, PlotType, PlotData, StatLayer, DataLayer, TimeSlicerEF, XAxisMixin, YLimMixin
from ._styles import Style, to_styles_dict


class UTSStat(LegendMixin, XAxisMixin, YLimMixin, EelFigure):
    """
    Plot statistics for a one-dimensional NDVar

    Parameters
    ----------
    y
        One or several dependent variable(s) (one-dimensional NDVar).
    x
        Model: specification of conditions which should be plotted separately.
    xax
        Make separate axes for each category in this categorial model.
    match
        Identifier for repeated measures data.
    sub
        Only use a subset of the data provided.
    data
        If a Dataset is specified, all data-objects can be specified as
        names of Dataset variables.
    main : func | None
        Measure for the central tendency (function that takes an ``axis``
        argument). The default is numpy.mean.
    error
        Measure of variability to plot. Examples:
        ``sem``: Standard error of the mean;
        ``2sem``: 2 standard error of the mean;
        ``ci``: 95% confidence interval;
        ``99%ci``: 99% confidence interval;
        ``sd``: standard deviation;
        ``all``: Show all traces;
        ``none``: No variability indication.
    within_subject_error
        Within-subject error bars (see Loftus & Masson, 1994; default is
        ``True`` for complete related measures designs, ``False`` otherwise).
    legend
        Matplotlib figure legend location argument or 'fig' to plot the
        legend in a separate figure.
    labels
        Alternative labels for legend as ``{cell: label}`` dictionary (preserves
        order).
    axtitle
        Title for the individual axes. The default is to show the names of the
        epochs, but only if multiple axes are plotted.
    xlabel
        X-axis label. By default the label is inferred from the data.
    ylabel
        Y-axis label. By default the label is inferred from the data.
    xticklabels
        Specify which axes should be annotated with x-axis tick labels.
        Use ``int`` for a single axis, a sequence of ``int`` for multiple
        specific axes, or one of ``'left' | 'bottom' | 'all' | 'none'``.
    yticklabels
        Specify which axes should be annotated with y-axis tick labels.
        Use ``int`` for a single axis, a sequence of ``int`` for multiple
        specific axes, or one of ``'left' | 'bottom' | 'all' | 'none'``.
    invy
        Invert the y axis (if ``bottom`` and/or ``top`` are specified explicitly
        they take precedence; an inverted y-axis can also be produced by
        specifying ``bottom > top``).
    bottom
        The lower end of the plot's y axis.
    top
        The upper end of the plot's y axis.
    case
        Dimension to treat as case (default is ``'case'``).
    xdim
        Dimension to plot along the x-axis (default is ``'time'``)
    xlim
        Initial x-axis view limits as ``(left, right)`` tuple or as ``length``
        scalar (default is the full x-axis in the data).
    clip
        Clip lines outside of axes (the default depends on whether ``frame`` is
        closed or open).
    color : matplotlib color
        Color if just a single category of data is plotted.
    colors
        Colors for the plots if multiple categories of data are plotted.
        **str**: A colormap name; Cells of ``x`` are mapped onto the colormap in
        regular intervals.
        **list**: A list of colors in the same sequence as ``x.cells``.
        **dict**: A dictionary mapping each cell in ``x`` to a color.
        Colors are specified as `matplotlib compatible color arguments
        <http://matplotlib.org/api/colors_api.html>`_.
    error_alpha
        Alpha of the error plot (default 0.3).
    mask : NDVar | {cell: NDVar}
        Mask certain time points. To control appearance of masked regions, set
        ``colors`` using :class:`plot.Style`.
    clusters
        Clusters to add to the plots. The clusters should be provided as
        Dataset, as stored in test results' :py:attr:`.clusters`.
    pmax
        Maximum p-value of clusters to plot as solid.
    ptrend
        Maximum p-value of clusters to plot as trend.
    tight
        Use matplotlib's tight_layout to expand all axes to fill the figure
        (default True)
    title
        Figure title.
    ...
        Also accepts :ref:`general-layout-parameters`.

    Examples
    --------
    - :ref:`exa-utsstat`

    Notes
    -----
    Navigation:
     - ``↑``: scroll up
     - ``↓``: scroll down
     - ``r``: y-axis zoom in (reduce y-axis range)
     - ``c``: y-axis zoom out (increase y-axis range)

    Examples
    --------
    Single :class:`NDVar` from a Dataset::

        plot.UTSStat('uts', 'A', 'B', match='rm', ds=ds)

    Multiple :class:`NDVar` in different axes

        plot.UTSStat(['uts1', 'uts2'], 'A', match='rm', ds=ds)

    Multiple :class:`NDVar` in a single axes::

        plot.UTSStat([['uts1', 'uts2']], match=True, ds=ds)

    """
    @deprecate_ds_arg
    def __init__(
            self,
            y: Union[NDVarArg, Sequence[NDVarArg]],
            x: CategorialArg = None,
            xax: CategorialArg = None,
            match: CategorialArg = None,
            sub: IndexArg = None,
            data: Dataset = None,
            main: Callable = np.mean,
            error: Union[str, Callable] = 'sem',
            within_subject_error: bool = None,
            legend: LegendArg = None,
            labels: Dict[CellArg, str] = None,
            axtitle: Union[bool, Sequence[str]] = True,
            xlabel: Union[bool, str] = True,
            ylabel: Union[bool, str] = True,
            xticklabels: Union[str, int, Sequence[int]] = 'bottom',
            yticklabels: Union[str, int, Sequence[int]] = 'left',
            invy: bool = False,
            bottom: float = None,
            top: float = None,
            case: str = None,
            xdim: str = None,
            xlim: Union[float, Tuple[float, float]] = None,
            clip: bool = None,
            colors: Dict[CellArg, Any] = None,
            error_alpha: float = 0.3,
            mask: Union[NDVar, Dict[CellArg, NDVar]] = None,
            clusters: Dataset = None,
            pmax: float = 0.05,
            ptrend: float = 0.1,
            **kwargs):
        if case and case != 'case':
            if x or xax or match:
                raise ValueError('x, xax and match cannot be specified with case')
            y = asndvar(y, sub, data)
            dimnames = y.get_dimnames(first=case)
            new_x = y.get_data(dimnames)
            dims = y.get_dims(dimnames[1:])
            y = NDVar(new_x, (Case, *dims))

        plot_data = PlotData.from_stats(y, x, xax, match, sub, data, (xdim,), colors, mask).for_plot(PlotType.LINE)
        xdim, = plot_data.dims

        layout = Layout(plot_data.plot_used, 2, 4, **kwargs)
        EelFigure.__init__(self, plot_data.frame_title, layout)
        if clip is None:
            clip = layout.frame is True

        # create plots
        self._plots = []
        legend_handles = {}
        ymax = ymin = None
        for ax, ax_data in zip(self.axes, plot_data):
            p = AxUTSStat(ax, ax_data, xdim, main, error, within_subject_error, clusters, pmax, ptrend, clip, error_alpha)
            self._plots.append(p)
            legend_handles.update(p.legend_handles)
            ymin = p.vmin if ymin is None else min(ymin, p.vmin)
            ymax = p.vmax if ymax is None else max(ymax, p.vmax)
        self._set_axtitle(axtitle, plot_data)

        # The legend should only display cells with distinct styles: remap legend handles to source cells
        alias_cells = {cell: style.alias for cell, style in plot_data.styles.items() if style.alias}
        if alias_cells:
            legend_handles = {alias_cells[cell]: handle for cell, handle in legend_handles.items()}

        # axes limits
        if top is not None:
            ymax = top
        if bottom is not None:
            ymin = bottom
        if invy:
            ymin, ymax = ymax, ymin
        for p in self._plots:
            p.set_ylim(ymin, ymax)

        self._configure_axis_data('y', plot_data.ct.y, ylabel, yticklabels)
        self._configure_axis_dim('x', plot_data.ct.y.get_dim(xdim), xlabel, xticklabels)
        XAxisMixin._init_with_data(self, ((plot_data.ct.y,),), xdim, xlim)
        YLimMixin.__init__(self, self._plots)
        LegendMixin.__init__(self, legend, legend_handles, labels, alt_sort=colors)
        self._update_ui_cluster_button()
        self._show()

    def _fill_toolbar(self, tb):
        from .._wxgui import wx

        btn = self._cluster_btn = wx.Button(tb, wx.ID_ABOUT, "Clusters")
        btn.Enable(False)
        tb.AddControl(btn)
        btn.Bind(wx.EVT_BUTTON, self._on_show_cluster_info)

        LegendMixin._fill_toolbar(self, tb)

    def _on_show_cluster_info(self, event):
        from .._wxgui import show_text_dialog

        if len(self._plots) == 1:
            clusters = self._plots[0].cluster_plt.clusters
            all_clusters = None
            all_plots_same = True
        else:
            all_clusters = [p.cluster_plt.clusters is None for p in self._plots]
            clusters = all_clusters[0]
            if all(c is clusters for c in all_clusters[1:]):
                all_plots_same = True
            else:
                all_plots_same = False

        if all_plots_same:
            info = str(clusters)
        else:
            info = []
            for i, clusters in enumerate(all_clusters):
                if clusters is None:
                    continue
                title = "Axes %i" % i
                info.append(title)
                info.append('\n')
                info.append('-' * len(title))
                info.append(str(clusters))
            info = '\n'.join(info)

        show_text_dialog(self._frame, info, "Clusters")

    def _update_ui_cluster_button(self):
        if hasattr(self, '_cluster_btn'):
            enable = not all(p.cluster_plt.clusters is None for p in self._plots)
            self._cluster_btn.Enable(enable)

    def set_clusters(
            self,
            clusters: Union[Dataset, None],
            pmax: float = 0.05,
            ptrend: float = None,
            color: Any = '.7',
            ax: int = None,
            y: Union[float, Dict] = None,
            dy: float = None,
            **kwargs,
    ):
        """Add clusters from a cluster test to the plot.

        Parameters
        ----------
        clusters
            A dataset with the clusters to visualize.
            Uses the format of clusters stored in test results'
            :py:attr:`.clusters`.
            This should include at least ``tstart`` and ``tstop`` columns.
            It can also optionally contain an ``effect`` column to use different
            ``color`` and ``y`` levels for different types of clusters.
            Use ``None`` to remove the clusters plotted on a given axis.
        pmax
            Only plot clusters with ``p <= pmax``.
        ptrend
            Maximum p-value of clusters to plot as trend.
        color : matplotlib color | dict
            Color for the clusters, or a ``{effect: color}`` dictionary.
        ax : None | int
            Index of the axes to which the clusters are to be added. If None,
            add the clusters to all axes.
        y
            Visualize clusters as horizontal bars at specific ``y`` levels in
            the plot (default is shaded boxes spanning the whole y-axis).
            As with ``color``, this can be an ``{effect: y}`` dictionary.
        dy
            Height of the bars when using ``y``.
        ...
            Additional arguments for the matplotlib artists (e.g., ``zorder``).

        Notes
        -----
        Previously added clusters will be removed.
        """
        axes = range(len(self.axes)) if ax is None else [ax]

        # update plots
        for ax in axes:
            p = self._plots[ax].cluster_plt
            p.kwargs = kwargs
            p.set_clusters(clusters, False)
            p.set_color(color, False)
            p.set_y(y, dy, False)
            p.set_pmax(pmax, ptrend)
        self.draw()

        self._update_ui_cluster_button()


class UTS(TimeSlicerEF, LegendMixin, YLimMixin, XAxisMixin, EelFigure):
    """Value by time plot for uniform time series (UTS) data

    Parameters
    ----------
    y : (list of) NDVar
        UTS data to plot. For a list (``[a, b]``) each item will be plotted
        separately. To combine multiple UTSs in a common axes, use a nested
        list (``[[a, b]]`` would plot ``a`` and ``b`` together in a common axes).
    xax : categorial
        Make separate axes for each category in this categorial model.
    axtitle : bool | sequence of str
        Title for the individual axes. The default is to show the names of the
        epochs, but only if multiple axes are plotted.
    data : Dataset
        If a Dataset is specified, all data-objects can be specified as
        names of Dataset variables.
    sub
        Specify a subset of the data.
    xlabel
        X-axis label. By default the label is inferred from the data.
    ylabel
        Y-axis label. By default the label is inferred from the data.
    xticklabels
        Specify which axes should be annotated with x-axis tick labels.
        Use ``int`` for a single axis, a sequence of ``int`` for multiple
        specific axes, or one of ``'left' | 'bottom' | 'all' | 'none'``.
    yticklabels
        Specify which axes should be annotated with y-axis tick labels.
        Use ``int`` for a single axis, a sequence of ``int`` for multiple
        specific axes, or one of ``'left' | 'bottom' | 'all' | 'none'``.
    bottom
        The lower end of the plot's y axis.
    top
        The upper end of the plot's y axis.
    legend
        Matplotlib figure legend location argument or 'fig' to plot the
        legend in a separate figure.
    labels
        Alternative labels for legend as ``{cell: label}`` dictionary (preserves
        order).
    xlim
        Initial x-axis view limits as ``(left, right)`` tuple or as ``length``
        scalar (default is the full x-axis in the data).
    clip
        Clip lines outside of axes (the default depends on whether ``frame`` is
        closed or open).
    colors
        Dictionary mapping ``y`` names to color, or a single color to use for
        all lines.
    stem
        Plot as stem-plot (default is a line-plot).
    tight : bool
        Use matplotlib's tight_layout to expand all axes to fill the figure
        (default True)
    ...
        Also accepts :ref:`general-layout-parameters`.

    Notes
    -----
    Navigation:
     - ``↑``: scroll up
     - ``↓``: scroll down
     - ``←``: scroll left
     - ``→``: scroll right
     - ``home``: scroll to beginning
     - ``end``: scroll to end
     - ``f``: x-axis zoom in (reduce x axis range)
     - ``d``: x-axis zoom out (increase x axis range)
     - ``r``: y-axis zoom in (reduce y-axis range)
     - ``c``: y-axis zoom out (increase y-axis range)
    """
    @deprecate_ds_arg
    def __init__(
            self,
            y: Union[NDVarArg, Sequence, NDTest],
            xax: CategorialArg = None,
            axtitle: Union[bool, Sequence[str]] = True,
            data: Dataset = None,
            sub: IndexArg = None,
            xlabel: Union[bool, str] = True,
            ylabel: Union[bool, str] = True,
            xticklabels: Union[str, int, Sequence[int]] = 'bottom',
            yticklabels: Union[str, int, Sequence[int]] = 'left',
            bottom: float = None,
            top: float = None,
            legend: LegendArg = None,
            labels: Dict[CellArg, str] = None,
            xlim: Union[float, Tuple[float, float]] = None,
            clip: bool = None,
            colors: Union[Any, dict] = None,
            color: Any = None,
            stem: bool = False,
            **kwargs):
        plot_data = PlotData.from_args(y, (None,), xax, data, sub).for_plot(PlotType.LINE)
        xdim = plot_data.dims[0]
        layout = Layout(plot_data.plot_used, 2, 4, **kwargs)
        EelFigure.__init__(self, plot_data.frame_title, layout)
        self._set_axtitle(axtitle, plot_data)
        self._configure_axis_dim('x', xdim, xlabel, xticklabels, data=plot_data.data)
        self._configure_axis_data('y', plot_data, ylabel, yticklabels)
        if clip is None:
            clip = layout.frame is True

        self.plots = []
        legend_handles = {}
        vlims = _base.find_fig_vlims(plot_data.data, top, bottom)

        n_colors = max(map(len, plot_data.data))
        if colors is None:
            if color is None:
                styles = [Style._coerce(color) for color in oneway_colors(n_colors)]
            else:
                styles = [Style._coerce(color)] * n_colors
        elif isinstance(colors, dict):
            styles = to_styles_dict(colors)
        else:
            styles = (Style._coerce(colors),) * n_colors

        for ax, layers in zip(self.axes, plot_data):
            h = AxUTS(ax, layers, xdim, vlims, styles, stem, clip)
            self.plots.append(h)
            legend_handles.update(h.legend_handles)

        self.epochs = plot_data.data
        XAxisMixin._init_with_data(self, plot_data.data, xdim, xlim)
        YLimMixin.__init__(self, self.plots)
        LegendMixin.__init__(self, legend, legend_handles, labels)
        TimeSlicerEF.__init__(self, xdim, plot_data.time_dim)
        self._show()

    def _fill_toolbar(self, tb):
        LegendMixin._fill_toolbar(self, tb)


class AxUTSStat:

    def __init__(
            self,
            ax: matplotlib.axes.Axes,
            data: AxisData,
            xdim: str,
            main: Callable,
            error: Union[str, Callable],
            within_subject_error: bool,
            clusters,
            pmax,
            ptrend,
            clip: bool,
            error_alpha,
    ):
        # stat plots
        self.ax = ax
        self.stat_plots = []
        self.legend_handles = {}

        for layer in data:
            plt = PltUTSStat(ax, layer, xdim, main, error, within_subject_error, clip, error_alpha)
            self.stat_plots.append(plt)
            if plt.main is not None:
                self.legend_handles[layer.style_key] = plt.main[0]

        # cluster plot
        self.cluster_plt = PltUTSClusters(ax, clusters, pmax, ptrend)

        # format y axis
        ax.autoscale(True, 'y')
        ax.autoscale(False, 'x')
        self.vmin, self.vmax = self.ax.get_ylim()

    @property
    def title(self):
        return self.ax.get_title()

    def set_ylim(self, vmin, vmax):
        self.ax.set_ylim(vmin, vmax)
        self.vmin, self.vmax = self.ax.get_ylim()


class AxUTS:

    def __init__(
            self,
            ax,
            axis_data: AxisData,
            xdim: str,
            vlims,
            styles: Union[Sequence, Dict],
            stem: bool,
            clip: bool = True,
    ):
        vmin, vmax = _base.find_uts_ax_vlim(axis_data.ndvars, vlims)
        if isinstance(styles, dict):
            styles = [styles[layer.y.name] for layer in axis_data]

        self.legend_handles = {}
        for layer, style in zip(axis_data, styles):
            p = PltUTS(ax, layer, xdim, style, stem, clip)
            self.legend_handles[longname(layer.y)] = p.plot_handle
            contours = layer.y.info.get('contours', None)
            if contours:
                for v, c in contours.items():
                    if v in contours:
                        continue
                    contours[v] = ax.axhline(v, color=c)

        self.ax = ax
        self.set_ylim(vmin, vmax)

    def set_ylim(self, vmin, vmax):
        self.ax.set_ylim(vmin, vmax)
        self.vmin, self.vmax = self.ax.get_ylim()


class PltUTS:

    def __init__(
            self,
            ax: matplotlib.axes.Axes,
            layer: DataLayer,
            xdim: str,
            style: Style,
            stem: bool = False,
            clip: bool = True,
    ):
        y = layer.y.get_data((xdim,))
        x = layer.y.get_dim(xdim)._axis_data()
        label = longname(layer.y)
        if stem:
            nonzero = y != 0
            nonzero[0] = True
            nonzero[-1] = True
            color = matplotlib.colors.to_hex(style.color)
            self.plot_handle = ax.stem(x[nonzero], y[nonzero], bottom=0, linefmt=color, markerfmt=' ', basefmt='#808080', label=label)
        else:
            kwargs = layer.plot_args(style.line_args)
            self.plot_handle = ax.plot(x, y, label=label, clip_on=clip, **kwargs)[0]

        for y, kwa in _base.find_uts_hlines(layer.y):
            ax.axhline(y, **kwa)


class PltUTSClusters:
    """UTS cluster plot"""
    def __init__(
            self,
            ax: matplotlib.axes.Axes,
            clusters: Dataset,  # 'tstart', 'tstop', 'p', 'effect'
            pmax: float,
            ptrend: float,
            style: Style = None,
    ):
        self.pmax = pmax
        self.ptrend = ptrend
        self.h = []
        self.ax = ax
        self.clusters = clusters
        self.style = style
        self.y = None
        self.dy = None
        self.kwargs = {}
        self.update()

    def set_clusters(self, clusters, update=True):
        self.clusters = clusters
        if update:
            self.update()

    def set_color(self, color, update=True):
        if isinstance(color, dict):
            self.style = to_styles_dict(color)
        else:
            self.style = Style._coerce(color)
        if update:
            self.update()

    def set_pmax(self, pmax, ptrend, update=True):
        self.pmax = pmax
        self.ptrend = ptrend
        if update:
            self.update()

    def set_y(self, y, dy, update=True):
        self.y = y
        self.dy = dy
        if update:
            self.update()

    def update(self):
        h = self.h
        while len(h):
            h.pop().remove()

        clusters = self.clusters
        if clusters is None:
            return

        if self.dy is None:
            bottom, top = self.ax.get_ylim()
            dy = (top - bottom) / 40.
        else:
            dy = self.dy

        p_include = self.ptrend or self.pmax
        for cluster in clusters.itercases():
            if 'p' in cluster:
                p = cluster['p']
                if p > p_include:
                    continue
                alpha = 0.5 if p < self.pmax else 0.2
            else:
                alpha = 0.5

            tstart = cluster['tstart']
            tstop = cluster['tstop']
            effect = cluster.get('effect')
            y = self.y[effect] if isinstance(self.y, dict) else self.y
            style = self.style[effect] if isinstance(self.style, dict) else self.style
            if y is None:
                kwargs = {'zorder': -10, **self.kwargs}
                if style:
                    kwargs = {**style.patch_args, **kwargs}
                h = self.ax.axvspan(tstart, tstop, fill=True, alpha=alpha, **kwargs)
            else:
                if style:
                    kwargs = {**style.patch_args, **self.kwargs}
                else:
                    kwargs = self.kwargs
                h = matplotlib.patches.Rectangle((tstart, y - dy / 2.), tstop - tstart, dy, linewidth=0, **kwargs)
                self.ax.add_patch(h)
            self.h.append(h)


class PltUTSStat:

    def __init__(
            self,
            ax: matplotlib.axes.Axes,
            layer: StatLayer,
            xdim: str,
            main: Callable,
            error: Union[str, Callable],
            within_subject_error: bool,
            clip: bool,
            error_alpha: float,
    ):
        # zorder defaults: 1 for patches, 2 for lines
        label = cellname(layer.cell)
        x = layer.y.get_dim(xdim)._axis_data()
        # plot main
        if callable(main):
            y_main = layer.get_statistic(main)
            kwargs = layer.style.line_args
            if error == 'all':
                lw = kwargs.get('linewidth') or matplotlib.rcParams['lines.linewidth']
                kwargs = {**kwargs, 'linewidth': lw * 2}
            self.main = ax.plot(x, y_main, label=label, clip_on=clip, **kwargs)
        elif error == 'all':
            self.main = y_main = None
        else:
            raise TypeError(f"{main=}")

        # plot error
        if error == 'all':
            y_all = layer.y.get_data((xdim, 'case'))
            self.error = ax.plot(x, y_all, alpha=error_alpha, clip_on=clip, **layer.style.line_args)
        elif error and error != 'none' and len(layer.y.get_dim('case')) > 1:
            if callable(error):
                dev_data = layer.get_statistic(error)
            else:
                dev_data = layer.get_dispersion(error, within_subject_error)
            lower = y_main - dev_data
            upper = y_main + dev_data
            self.error = ax.fill_between(x, lower, upper, linewidth=0, clip_on=clip, **layer.style.fill_args(error_alpha))
        else:
            self.error = None



import numpy as np
from eelbrain import *
import eelbrain
from eelbrain._stats import opt

ds = datasets.get_uts(True)
y = ds['utsnd'].x
n_cases = len(y)
n_tests = np.product(y.shape[1:])
y = y.reshape((n_cases, n_tests))
m = ds.eval("A*B*rm")
p = m._parametrize()
fmap = np.empty((3, n_tests))
e_ms = eelbrain._stats.glm._hopkins_ems_array(m)

print("timeit -n1000 opt.anova_full_fmaps(y, p.x, p.projector, fmap, m._effect_to_beta, e_ms)")



"Tools for loading data from text files."
import csv
import gzip
from pathlib import Path
import re
from numbers import Number
from typing import Sequence, Tuple, Union

import numpy as np

from .._types import PathArg
from .._utils import ui
from .._utils.parse import FLOAT_NAN_PATTERN, PERCENT_PATTERN
from .. import _data_obj as _data

__all__ = ('tsv', 'var')


def to_num(v: str):
    if v.isdigit():
        return int(v)
    else:
        return float(v)


def tsv(
        path: PathArg = None,
        names: Union[Sequence[str], bool] = True,
        types: Union[str, dict] = None,
        delimiter: str = 'auto',
        skiprows: int = 0,
        start_tag: str = None,
        ignore_missing: bool = False,
        empty: Union[str, float] = None,
        random: Union[str, Sequence[str]] = None,
        strip: bool = False,
        encoding: str = None,
        comment: str = '^#',
        percent_to_proportion: bool = False,
        **fmtparams,
) -> _data.Dataset:
    r"""Load a :class:`Dataset` from a text file.

    Parameters
    ----------
    path : str
        Path to the file (if omitted, use a system file dialog). Files ending
        in ``*.gz`` are automatically decompressed.
    names : Sequence of str | bool
        Column/variable names.

        * ``True`` (default): look for names on the first line of the file
        * ``['name1', ...]`` use these names
        * ``False``: use "v1", "v2", ...
    types : str | dict
        Column data types::

         - 'a': determine automatically
         - 'f': Factor
         - 'v': Var
         - 'b': boolean Var

        Specified either as string, with one type per columnc (e.g.
        ``'ffvva'``) or as ``{column_name: data_type}`` dictionary (e.g.
        ``{'participant': 'f'}``); unspecified columns default to ``'a'``.
    delimiter : str
        Value delimiting cells in the input file (default depends on ``path``:
        ``','`` if the extension is ``'.csv'``, otherwise ``'\t'``).
    skiprows : int
        Skip so many rows at the beginning of the file (for tsv files with
        headers). Column names are expected to come after the skipped rows.
        ``skiprows`` is applied after ``start_tag``.
    start_tag : str
        Alternative way to skip header rows. The table is assumed to start
        on the line following the last line in the file that starts with
        ``start_tag``.
    ignore_missing : bool
        Ignore rows with missing values (i.e., lines that contain fewer
        ``delimiter`` than the others; by default this raises an IOError). For
        rows with missing values, ``NaN`` is substituted for numerical and
        ``""`` for categorial variables.
    empty : number | 'nan'
        For numerical variables, substitute this value for empty entries. For
        example, if a column in a file contains ``['5', '3', '']``, this is read
        by default as ``Factor(['5', '3', ''])``.
        With ``empty=0``, it is read as ``Var([5, 3, 0])``.
        With ``empty='nan'``, it is read as ``Var([5, 3, nan])``.
    random : str | sequence of str
        Names of the columns that should be assigned as random factor.
    strip
        Strip white-space from all categorial variables.
    encoding
        Text file encoding (see :func:`open`).
    comment
        Regular expression for lines to skip (default is lines starting with
        ``#``).
    percent_to_proportion
        Convert columns that end in the percent sign into a float reflecting
        proportion (i.e. strip the percent sign and divide the number by 100).
    **fmtparams
        Further formatting parameters for :func:`csv.reader`. For example, a
        fixed-width column file can be loaded with ``skipinitialspace=True``
        (as long as there are no empty cells).

    See Also
    --------
    Dataset.save_txt : Save :class:`Dataset` in a text file
    """
    if path is None:
        path = ui.ask_file("Load TSV", "Select tsv file to import as Dataset")
        if not path:
            return
    path = Path(path).expanduser()

    if isinstance(random, str):
        random = [random]
    elif random is None:
        random = []
    else:
        random = list(random)

    suffix = path.suffix.lower()
    if suffix == '.gz':
        open_ = gzip.open
        suffix = path.with_suffix('').suffix.lower()
    else:
        open_ = open

    if delimiter is None:  # legacy option
        delimiter = ' '
        fmtparams['skipinitialspace'] = True
    elif delimiter == 'auto':
        if suffix == '.csv':
            delimiter = ','
        else:
            delimiter = '\t'

    comment_pattern = re.compile(comment) if comment else None

    with open_(path, 'rt', encoding=encoding, newline='') as lines:
        if comment_pattern is not None:
            lines = (line for line in lines if not comment_pattern.match(line))
        reader = csv.reader(lines, delimiter=delimiter, **fmtparams)
        lines = list(reader)
    if lines[0][0].startswith('\ufeff'):
        raise IOError(f"First word invalid: {lines[0][0]!r}; file might be encoded with byte order mark, try opening with encoding='utf-8-sig' (see https://stackoverflow.com/a/17912811/166700)")

    # find start position
    if start_tag:
        start = 0
        for i, line in enumerate(lines, 1):
            if line[0].startswith(start_tag):
                start = i
        if start:
            lines = lines[start:]
    if skiprows:
        lines = lines[skiprows:]

    # determine column names
    if names is True:
        column_names = lines.pop(0)
    elif names:
        column_names = list(names)
    else:
        column_names = []

    # determine number of columns
    row_lens = set(len(row) for row in lines)
    if not ignore_missing and len(row_lens) > 1:
        raise IOError("Not all rows have same number of entries. Set ignore_missing to True in order to ignore this error.")
    n_columns = max(row_lens)

    # check/adjust column names
    if names is True:  # R write.table saves unnamed column with row names
        if len(column_names) == n_columns - 1:
            name = "row"
            while name in column_names:
                name += '_'
            column_names.insert(0, name)
    elif names:
        if len(column_names) > n_columns:
            raise IOError(f"{names=}: More names than columns ({len(names)=}, {n_columns=})")
    # fill in missing column names
    for column in range(len(column_names), n_columns):
        key = f'v{column}'
        while key in column_names:
            key += '_'
        column_names.append(key)

    # check random
    missing = [k for k in random if k not in column_names]
    if missing:
        raise ValueError(f"{random=} includes non-existent names: {', '.join(missing)}")

    # coerce types parameter
    if types is None:
        types_ = 'a' * n_columns
    elif isinstance(types, dict):
        types_ = [types.get(n, 'a') for n in column_names]
    elif isinstance(types, str):
        types_ = types
    elif isinstance(types, (list, tuple)):
        # backwards compatibility
        types_ = ['afv'[v] for v in types]
    else:
        raise TypeError(f'types={types!r}')

    # check types values
    if len(types_) != n_columns:
        raise ValueError(f'{types=}: {len(types)} values for file with {n_columns} columns')
    elif set(types_).difference('afvb'):
        invalid = ', '.join(map(repr, set(types_).difference('afvb')))
        raise ValueError(f'{types=}: invalid values {invalid}')

    # find quotes (imply type 1)
    # quotes = "'\""
    data = np.empty((len(lines), n_columns), object)
    for r, line in enumerate(lines):
        for c, v in enumerate(line):
            # for str_del in quotes:
            #     if len(v) > 0 and v[0] == str_del:
            #         v = v.strip(str_del)
            #         types_[c] = 'f'
            data[r, c] = v

    # convert values to data-objects
    float_pattern = re.compile(FLOAT_NAN_PATTERN)
    percent_pattern = re.compile(PERCENT_PATTERN)
    ds = _data.Dataset(name=path.name)
    np_vars = {
        'NA': np.nan,
        'na': np.nan,
        'NaN': np.nan,
        'nan': np.nan,
        'NAN': np.nan,
        None: np.nan,
    }
    if empty is not None:
        if isinstance(empty, str):
            np_vars[''] = to_num(empty)
        elif isinstance(empty, Number):
            np_vars[''] = empty
        else:
            raise TypeError(f'{empty=}')
    bool_dict = {'True': True, 'False': False, None: False}
    keys = {}
    for name, values, type_ in zip(column_names, data.T, types_):
        # infer type
        if type_ in 'fvb%':
            pass
        elif all(v in bool_dict for v in values):
            type_ = 'b'
        elif empty is not None:
            if all(v in (None, '') or float_pattern.match(v) for v in values):
                type_ = 'v'
            elif percent_to_proportion and all(v in (None, '') or percent_pattern.match(v) for v in values):
                type_ = '%'
            else:
                type_ = 'f'
        elif all(v is None or float_pattern.match(v) for v in values):
            type_ = 'v'
        elif all(v is None or percent_pattern.match(v) for v in values):
            type_ = '%'
        else:
            type_ = 'f'

        # substitute values
        if type_ == '%':
            values = [v.rstrip().rstrip('%') for v in values]
        if type_ in 'v%':
            values = [np_vars[v] if v in np_vars else to_num(v) for v in values]
        elif type_ == 'b':
            values = [bool_dict[v] for v in values]

        # create data-object
        if type_ == 'f':
            f_random = name in random
            d_obj = _data.Factor(values, labels={None: ''}, name=name, random=f_random)
            if strip:
                if any(cell.strip() != cell for cell in d_obj.cells):
                    d_obj.update_labels({cell: cell.strip() for cell in d_obj.cells})
        elif name in random:
            raise ValueError(f"{random=}: {name} is not categorial")
        else:
            d_obj = _data.Var(values, name)
            if type_ == '%':
                d_obj /= 100
        key = _data.Dataset.as_key(name)
        keys[name] = key
        ds[key] = d_obj

    if any(k != v for k, v in keys.items()):
        ds.info['keys'] = keys

    return ds


def var(path: PathArg = None, name: str = None):
    """
    Load a :class:`Var` object from a text file by splitting at white-spaces.

    Parameters
    ----------
    path : str
        Source file. If not specified, a system file dialog is opened.
    name : str
        Name for the Var.
    """
    if path is None:
        path = ui.ask_file("Load Var File", "Select Var File")
        if not path:
            return

    with open(path) as file:
        lines = file.read().split()
    is_bool = all(line in ['True', 'False'] for line in lines)

    if is_bool:
        x = np.genfromtxt(path, dtype=bool)
    else:
        x = np.loadtxt(path)

    return _data.Var(x, name)


def write_adjacency(
        adjacency: Union[Sequence[Tuple[str, str]], Sequence[Tuple[int, int]]],
        filename: PathArg = None,
):
    """Save adjacency graph as text file"""
    text = '\n'.join([':'.join(map(str, pair)) for pair in adjacency])
    if filename is None:
        msg = "Save adjacency..."
        filename = ui.ask_saveas(msg, msg, [("Text files", "*.txt")])
    Path(filename).write_text(text)


def read_adjacency(
        filename: PathArg = None,
        is_int: bool = False,
):
    if filename is None:
        filename = ui.ask_file("Load Adjacency", "Select text file with adjacency graph", [("Text files", "*.txt")])
        if filename is None:
            return
    text = Path(filename).read_text()
    pairs = [pair.split(':') for pair in text.splitlines()]
    if is_int:
        pairs = [[int(i) for i in pair] for pair in pairs]
    return [tuple(pair) for pair in pairs]



import wx
from eelbrain import gui
from eelbrain.testing import gui_test, TempDir
from eelbrain._io.tests.test_stc_dataset import _create_fake_files


@gui_test
def test_load_stcs():
    tmp = TempDir()
    _create_fake_files(tmp)
    frame = gui.load_stcs()
    assert frame.dir_ctl.GetPath() == ""
    frame.dir_ctl.SetPath(tmp)
    evt = wx.FileDirPickerEvent()
    evt.SetPath(tmp)
    frame.OnDirChange(evt)
    assert frame.loader is not None
    assert len(frame.factor_name_ctrls) == 2
    assert "verb" in frame.loader.levels[1]
    frame.Close()


@gui_test
def test_failed_load():
    tmp = TempDir()
    frame = gui.load_stcs()
    frame.dir_ctl.SetPath(tmp)
    evt = wx.FileDirPickerEvent()
    evt.SetPath(tmp)
    frame.OnDirChange(evt)
    assert "No .stc" in frame.status.GetStatusText()
    frame.Close()



"""Variables

With multiple sessions, the same variable name might have multiple definitions::

variables = {
    'name': [Var(session='1'), Var(session='2')],
    'name2': Var(session='2'),
}

"""
from fnmatch import fnmatch as fnmatch_func
from typing import Dict, Sequence, Tuple, Union

import numpy as np

from .._data_obj import Factor, Var, assert_is_legal_dataset_key
from .._utils.numpy_utils import INT_TYPES
from .._utils.parse import find_variables
from .definitions import DefinitionError


# Some event columns are reserved for Eelbrain
RESERVED_VAR_KEYS = ('subject', 'task', 'visit')


def as_vardef_var(v):
    "Coerce ds.eval() output for use as variable"
    if isinstance(v, np.ndarray):
        if v.dtype.kind == 'b':
            return Var(v.astype(int))
        return Var(v)
    return v


class VarDef:
    _pickle_args = ('task',)

    def __init__(self, task):
        self.task = task

    def __getstate__(self):
        return {k: getattr(self, k) for k in self._pickle_args}

    def __setstate__(self, state):
        for k in self._pickle_args:
            setattr(self, k, state[k])

    @property
    def _eq_args(self):
        raise NotImplementedError

    def __eq__(self, other):
        return isinstance(other, self.__class__) and other._eq_args == self._eq_args

    def apply(self, ds, e):
        raise NotImplementedError

    def input_vars(self):
        raise NotImplementedError


class EvalVar(VarDef):
    """Variable based on evaluating a statement

    Parameters
    ----------
    code
        Statement to evaluate.
    task
        Only apply the variable to events from this task.

    See Also
    --------
    MneExperiment.variables
    """
    _pickle_args = ('task', 'code')

    def __init__(self, code: str, task: str = None):
        super(EvalVar, self).__init__(task)
        assert isinstance(code, str)
        self.code = code

    def __repr__(self):
        return "EvalVar(%r)" % self.code

    @property
    def _eq_args(self):
        return self.code,

    def apply(self, ds, e):
        return as_vardef_var(ds.eval(self.code))

    def input_vars(self):
        return find_variables(self.code)


class LabelVar(VarDef):
    """Variable assigning labels to values

    Parameters
    ----------
    source
        Variable supplying the values (e.g., ``"trigger"``).
    codes
        Mapping values in ``source`` to values in the new variable. The type
        of the values determines whether the output is a :class:`Factor`
        (:class:`str` values) or a :class:`Var` (numerical values).
    default
        Label for values not in ``codes``. By default, this is ``''`` for
        categorial and 0 for numerical output. Set to ``False`` to pass through
        unlabeled input values.
    task
        Only apply the variable to events from this task.
    fnmatch
        Treat keys in ``codes`` as :mod:`fnmatch` patterns.

    See Also
    --------
    MneExperiment.variables
    """
    _pickle_args = ('task', 'source', 'codes', 'labels', 'is_factor', 'default', 'fnmatch')

    def __init__(
            self,
            source: str,
            codes: Dict[Union[str, float, Tuple[str, ...], Tuple[float, ...]], Union[str, float]],
            default: Union[bool, str, float] = True,
            task: str = None,
            fnmatch: bool = False,
    ):
        super(LabelVar, self).__init__(task)
        self.source = source
        self.codes = codes
        self.labels = {}
        is_factor = None
        for key, v in codes.items():
            if is_factor is None:
                is_factor = isinstance(v, str)
            elif isinstance(v, str) != is_factor:
                raise DefinitionError(f"LabelVar with {codes=}: value type inconsistent, need all or none to be str")

            if isinstance(key, tuple):
                for k in key:
                    self.labels[k] = v
            else:
                self.labels[key] = v
        self.is_factor = is_factor
        if default is True:
            default = '' if is_factor else 0
        elif default is not None:
            if isinstance(default, str) != is_factor:
                raise TypeError(f"{default=}")
        self.default = default
        self.fnmatch = fnmatch

    def __repr__(self):
        return f"{self.__class__.__name__}({self.source!r}, {self.codes})"

    @property
    def _eq_args(self):
        return self.source, self.labels, self.default, self.fnmatch

    def apply(self, ds, e):
        source = ds.eval(self.source)
        if self.fnmatch:
            labels = {}
            for value in source.cells:
                for pattern, target in self.labels.items():
                    if fnmatch_func(value, pattern):
                        labels[value] = target
        else:
            labels = self.labels
        if self.is_factor:
            return Factor(source, labels=labels, default=self.default)
        else:
            return Var.from_dict(source, labels, default=self.default)

    def input_vars(self):
        return find_variables(self.source)


class GroupVar(VarDef):
    """Group membership for each subject

    Parameters
    ----------
    groups
        Groups to label. A sequence of group names to label each subject with
        the group it belongs to (subjects can't be members of more than one
        group). Alternatively, a ``{group: label}`` dictionary can be used to
        assign a different label based on group membership.
    task
        Only apply the variable to events from this task.

    See Also
    --------
    MneExperiment.variables

    Examples
    --------
    Assuming an experiment that defines two groups, ``'patient'`` and
    ``'control'``, these groups could be labeled with::

        GroupVar(['patient', 'control'])

    """
    _pickle_args = ('task', 'groups')

    def __init__(
            self,
            groups: Union[Sequence[str], Dict[str, str]],
            task: str = None,
    ):
        super(GroupVar, self).__init__(task)
        self.groups = groups

    def __repr__(self):
        return f"GroupVar({self.groups!r})"

    @property
    def _eq_args(self):
        return self.groups,

    def apply(self, ds, e):
        return e.label_groups(ds['subject'], self.groups)

    @classmethod
    def from_string(cls, string):
        groups = {}
        for item in string.split(','):
            if ':' in item:
                src, dst = map(str.strip, item.split(':'))
            else:
                src = dst = item.strip()
            groups[src] = dst
        if all(k == v for k, v in groups.items()):
            groups = tuple(sorted(groups))
        return cls(groups)

    def input_vars(self):
        return ()


def parse_named_vardef(string):
    if '=' not in string:
        raise DefinitionError(f"variable {string!r}: needs '='")
    name, vdef = string.split('=', 1)
    return name.strip(), parse_vardef(vdef)


def parse_vardef(string):
    string = string.strip()
    if string.startswith('group:'):
        return GroupVar.from_string(string[6:])
    else:
        return EvalVar(string)


class Variables:
    """Set of variable definitions

    Parameters
    ----------
    arg : str | tuple | dict
        The ``vars`` argument.
    """
    def __init__(self, arg: dict):
        if arg is None:
            arg = ()
        elif isinstance(arg, str):
            arg = (arg,)
        elif isinstance(arg, dict):
            arg = arg.items()
        elif not isinstance(arg, (tuple, list)):
            raise TypeError(f"vars={arg!r}")

        self.vars = {}
        for item in arg:
            if isinstance(item, str):
                name, vdef = parse_named_vardef(item)
            else:
                name, vdef = item
                if isinstance(vdef, str):
                    vdef = parse_vardef(vdef)
                elif isinstance(vdef, VarDef):
                    pass
                elif isinstance(vdef, dict):
                    if 'default' in vdef:
                        vdef = vdef.copy()
                        default = vdef.pop('default')
                    else:
                        default = True
                    vdef = LabelVar('trigger', vdef, default)
                elif isinstance(vdef, tuple):
                    vdef = LabelVar(*vdef)
                else:
                    raise DefinitionError(f"Variable {name!r}: {vdef!r}")

            assert_is_legal_dataset_key(name)
            if name in RESERVED_VAR_KEYS:
                raise DefinitionError(f"Variable {name!r}: reserved name")
            self.vars[name] = vdef

    def __getstate__(self):
        return {'vars': self.vars}

    def __setstate__(self, state):
        self.vars = state['vars']

    def _check_trigger_vars(self):
        for key, var in self.vars.items():
            if isinstance(var, LabelVar) and var.source == 'trigger':
                if not all(isinstance(v, INT_TYPES) for v in var.labels):
                    raise DefinitionError(f"Variable {key!r}: {var} codes must be integers")

    def __repr__(self):
        return '\n'.join(["Variables(", *(f'    {k!r}: {v},' for k, v in self.vars.items()), ')'])

    def __eq__(self, other):
        return isinstance(other, Variables) and other.vars == self.vars

    def apply(self, ds, e, group_only=False):
        task = ds.info.get('task', None)
        for name, vdef in self.vars.items():
            if group_only and not isinstance(vdef, GroupVar):
                continue
            elif vdef.task is None or vdef.task == task:
                ds[name] = vdef.apply(ds, e)



# Author: Christian Brodbeck <christianbrodbeck@nyu.edu>
"""
.. _exa-cluster-based-mu:

T-test
======

.. currentmodule:: eelbrain

This example show a cluster-based permutation test for a simple design (two
conditions). The example uses simulated data meant to vaguely resemble data
from an N400 experiment (not intended as a physiologically realistic
simulation).

.. contents:: Contents
   :local:

"""
# sphinx_gallery_thumbnail_number = 3
from eelbrain import *

###############################################################################
# Simulated data
# --------------
# Each function call to :func:`datasets.simulate_erp` generates a dataset
# equivalent to an N400 experiment for one subject. 
# The ``seed`` argument determines the random noise that is added to the data.
ds = datasets.simulate_erp(seed=0)
print(ds.summary())

###############################################################################
# A singe trial of data:
p = plot.TopoButterfly('eeg[0]', data=ds, t=0.400)

###############################################################################
# The :meth:`Dataset.aggregate` method computes condition averages when sorting
# the data into conditions of interest. In our case, the ``predictability`` variable
# specified conditions ``'high'`` and ``'low'`` cloze:
print(ds.aggregate('predictability'))

###############################################################################
# Group level data
# ----------------
# This loop simulates a multi-subject experiment. 
# It generates data and collects condition averages for 10 virtual subjects. 
# For group level analysis, the collected data are combined in a :class:`Dataset`:
dss = []
for subject in range(10):
    # generate data for one subject
    ds = datasets.simulate_erp(seed=subject)
    # average across trials to get condition means
    ds_agg = ds.aggregate('predictability')
    # add the subject name as variable
    ds_agg[:, 'subject'] = f'S{subject:02}'
    dss.append(ds_agg)

ds = combine(dss)
# make subject a random factor (to treat it as random effect for ANOVA)
ds['subject'].random = True
ds.head()

###############################################################################
# Re-reference the EEG data (i.e., subtract the mean of the two mastoid channels):
ds['eeg'] -= ds['eeg'].mean(sensor=['M1', 'M2'])

###############################################################################
# Spatio-temporal cluster based test
# ----------------------------------
# Cluster-based tests are based on identifying clusters of meaningful effects, i.e., 
# groups of adjacent sensors that show the same effect (see :mod:`testnd` for references).
# In order to find clusters, the algorithm needs to know which channels are 
# neighbors. This information is refered to as the sensor adjacency (i.e., which sensors 
# are connected). The adjacency graph can be visualized to confirm that it is set correctly.
p = plot.SensorMap(ds['eeg'], adjacency=True)

###############################################################################
# With the correct adjacency, we can now compute a cluster-based permutation test
# for a related measures *t*-test:
res = testnd.TTestRelated(
    'eeg', 'predictability', 'low', 'high', match='subject', data=ds,
    pmin=0.05,  # Use uncorrected p = 0.05 as threshold for forming clusters
    tstart=0.100,  # Find clusters in the time window from 100 ...
    tstop=0.600,  # ... to 600 ms
)

###############################################################################
# Plot the test result. The top two rows show the two condition averages. The
# bottom butterfly plot shows the difference with only significant regions in
# color. The corresponding topomap shows the topography at the marked time
# point, with the significant region circled (in an interactive environment,
# the mouse can be used to update the time point shown).
p = plot.TopoButterfly(res, clip='circle', t=0.400)

###############################################################################
# Generate a table with all significant clusters:
clusters = res.find_clusters(0.05)
clusters

###############################################################################
# Retrieve the cluster map using its ID and visualize the spatio-temporal
# extent of the cluster:
cluster_id = clusters[0, 'id']
cluster = res.cluster(cluster_id)
p = plot.TopoArray(cluster, interpolation='nearest', t=[0.350, 0.400, 0.450, None])
# plot the colorbar next to the right-most sensor plot
p_cb = p.plot_colorbar(right_of=p.axes[3])

###############################################################################
# Using a cluster as functional ROI
# ---------------------------------
# Often it is desirable to summarize values in a cluster. This is especially useful
# in more complex designs. For example, after finding a signficant interaction effect 
# in an ANOVA, one might want to follow up with a pairwise test of the value in the 
# cluster. This can often be achieved using binary masks based on the cluster. Using
# the cluster identified above, generate a binary mask:
mask = cluster != 0
p = plot.TopoArray(mask, cmap='Wistia', t=[0.350, 0.400, 0.450])

###############################################################################
# Such a spatio-temporal boolean mask can be used
# to extract the value in the cluster for each condition/participant.
# Since ``mask`` contains both time and sensor dimensions, using it with 
# the :meth:`NDVar.mean` method collapses across these dimensions and 
# returns a scalar for each case (i.e., for each condition/subject).
ds['cluster_mean'] = ds['eeg'].mean(mask)
p = plot.Barplot('cluster_mean', 'predictability', match='subject', data=ds, test=False)

###############################################################################
# Similarly, a mask consisting of a cluster of sensors can be used to 
# visualize the time course in that region of interest. A straight forward
# choice is to use all sensors that were part of the cluster (``mask``)
# at any point in time:
roi = mask.any('time')
p = plot.Topomap(roi, cmap='Wistia')

###############################################################################
# When using a mask that ony contains a sensor dimension (``roi``),
# :meth:`NDVar.mean` collapses across sensors and returns a value for each time
# point, i.e. the time course in sensors involved in the cluster:
ds['cluster_timecourse'] = ds['eeg'].mean(roi)
p = plot.UTSStat('cluster_timecourse', 'predictability', match='subject', data=ds, frame='t')
# mark the duration of the spatio-temporal cluster
p.set_clusters(clusters, y=0.25e-6)

###############################################################################
# Now visualize the cluster topography, marking significant sensors:
time_window = (clusters[0, 'tstart'], clusters[0, 'tstop'])
c1_topo = res.c1_mean.mean(time=time_window)
c0_topo = res.c0_mean.mean(time=time_window)
diff_topo = res.difference.mean(time=time_window)
p = plot.Topomap([c1_topo, c0_topo, diff_topo], axtitle=['Low cloze', 'High cloze', 'Low - high'], columns=3)
p.mark_sensors(roi, -1)

###############################################################################
# Temporal cluster based test
# ---------------------------
# Alternatively, if a spatial region of interest exists, a univariate time
# course can be extracted and submitted to a temporal cluster based test. For
# example, the N400 is typically expected to be strong at sensor ``Cz``:
ds['eeg_cz'] = ds['eeg'].sub(sensor='Cz')
res_timecoure = testnd.TTestRelated(
    'eeg_cz', 'predictability', 'low', 'high', match='subject', data=ds,
    pmin=0.05,  # Use uncorrected p = 0.05 as threshold for forming clusters
    tstart=0.100,  # Find clusters in the time window from 100 ...
    tstop=0.600,  # ... to 600 ms
)
clusters = res_timecoure.find_clusters(0.05)
clusters

p = plot.UTSStat('eeg_cz', 'predictability', match='subject', data=ds, frame='t')
p.set_clusters(clusters, y=0.25e-6)



# Author: Christian Brodbeck <christianbrodbeck@nyu.edu>
from eelbrain import datasets, boosting, epoch_impulse_predictor
from eelbrain.testing import hide_plots


@hide_plots
def test_datasplit():
    ds = datasets.get_uts()
    ds['x'] = epoch_impulse_predictor('uts', data=ds)

    trf = boosting('uts', 'x', 0, 0.100, data=ds, partitions=4)
    p = trf.splits.plot()
    p.close()

    trf = boosting('uts', 'x', 0, 0.100, data=ds, partitions=4, test=True)
    p = trf.splits.plot()
    p.close()



# Author: Christian Brodbeck <christianbrodbeck@nyu.edu>
import numpy
from numpy.testing import assert_equal

from eelbrain._ndvar._convolve import convolve_1d


def test_convolve_1d():
    # n_x = 1
    pad_0 = numpy.array([0], float)
    pad_1 = numpy.array([1], float)
    h = numpy.array([[1, 2]], float)
    x = numpy.array([[0, 1, 0, 0, 1.5, 0]])
    yt = numpy.array([0, 1, 2, 0, 1.5, 3])  # target
    n_times = x.shape[1]
    segments = numpy.array([[0, n_times]], numpy.int64)
    yp = numpy.empty(yt.shape)
    convolve_1d(h, x, pad_0, 0, segments, yp)
    assert_equal(yp, yt)
    # shift -1
    yt = numpy.array([1, 2, 0, 1.5, 3, 0])  # target
    convolve_1d(h, x, pad_0, -1, segments, yp)
    assert_equal(yp, yt)
    # shift +1
    yt = numpy.array([0, 0, 1, 2, 0, 1.5])  # target
    convolve_1d(h, x, pad_0, 1, segments, yp)
    assert_equal(yp, yt)
    # pad head
    yt = numpy.array([2, 1, 2, 0, 1.5, 3])  # target
    convolve_1d(h, x, pad_1, 0, segments, yp)
    assert_equal(yp, yt)
    # pad tail
    yt = numpy.array([1, 2, 0, 1.5, 3, 1])  # target
    convolve_1d(h, x, pad_1, -1, segments, yp)
    assert_equal(yp, yt)



# Author: Christian Brodbeck <christianbrodbeck@nyu.edu>
from itertools import product
import pickle
import logging
import pytest
import sys

import numpy as np
from numpy.testing import assert_array_equal, assert_allclose

import eelbrain
from eelbrain import Dataset, NDVar, Categorial, Scalar, UTS, Sensor, configure, datasets, test, testnd, set_log_level, cwt_morlet
from eelbrain._exceptions import WrongDimensionError, ZeroVarianceError
from eelbrain._stats.testnd import Adjacency, NDPermutationDistribution, label_clusters, _MergedTemporalClusterDist, find_peaks, VectorDifferenceIndependent
from eelbrain._utils.system import IS_WINDOWS
from eelbrain.fmtxt import asfmtext
from eelbrain.testing import assert_dataobj_equal, assert_dataset_equal, requires_mne_sample_data


def test_anova():
    "Test testnd.ANOVA()"
    ds = datasets.get_uts(True, nrm=True)

    testnd.ANOVA('utsnd', 'A*B', data=ds)
    for samples in (0, 2):
        logging.info("TEST:  samples=%r" % samples)
        testnd.ANOVA('utsnd', 'A*B', data=ds, samples=samples)
        testnd.ANOVA('utsnd', 'A*B', data=ds, samples=samples, pmin=0.05)
        res = testnd.ANOVA('utsnd', 'A*B', data=ds, samples=samples, tfce=True)
        assert res._plot_model() == 'A%B'
    asfmtext(res)

    res = testnd.ANOVA('utsnd', 'A*B*rm', match=False, data=ds, samples=0, pmin=0.05)
    assert repr(res) == "<ANOVA 'utsnd', 'A*B*rm', match=False, samples=0, pmin=0.05, 'A': 17 clusters, 'B': 20 clusters, 'A x B': 22 clusters>"
    assert res._plot_model() == 'A%B'
    res = testnd.ANOVA('utsnd', 'A*B*rm', data=ds, samples=2, pmin=0.05)
    assert res.match == 'rm'
    assert repr(res) == "<ANOVA 'utsnd', 'A*B*rm', match='rm', samples=2, pmin=0.05, 'A': 17 clusters, p < .001, 'B': 20 clusters, p < .001, 'A x B': 22 clusters, p < .001>"
    assert res._plot_model() == 'A%B'

    # persistence
    string = pickle.dumps(res, protocol=pickle.HIGHEST_PROTOCOL)
    res_ = pickle.loads(string)
    assert repr(res_) == repr(res)
    assert res_._plot_model() == 'A%B'

    # threshold-free
    res = testnd.ANOVA('utsnd', 'A*B*rm', data=ds, samples=10)
    assert res.match == 'rm'
    assert repr(res) == "<ANOVA 'utsnd', 'A*B*rm', match='rm', samples=10, 'A': p < .001, 'B': p < .001, 'A x B': p < .001>"
    assert 'A clusters' in res.clusters.info
    assert 'B clusters' in res.clusters.info
    assert 'A x B clusters' in res.clusters.info

    # no clusters
    res = testnd.ANOVA('uts', 'B', sub="A=='a1'", data=ds, samples=5, pmin=0.05, mintime=0.02)
    repr(res)
    assert 'v' in res.clusters
    assert 'p' in res.clusters
    assert res._plot_model() == 'B'

    # all effects with clusters
    res = testnd.ANOVA('uts', 'A*B*rm', match=False, data=ds, samples=5, pmin=0.05, tstart=0.1, mintime=0.02)
    assert set(res.clusters['effect'].cells) == set(res.effects)

    # some effects with clusters, some without
    res = testnd.ANOVA('uts', 'A*B*rm', data=ds, samples=5, pmin=0.05, tstart=0.37, mintime=0.02)
    assert res.match == 'rm'
    string = pickle.dumps(res, pickle.HIGHEST_PROTOCOL)
    res_ = pickle.loads(string)
    assert_dataobj_equal(res.clusters, res_.clusters)

    # test multi-effect results (with persistence)
    # UTS
    res = testnd.ANOVA('uts', 'A*B*rm', data=ds, samples=5)
    assert res.match == 'rm'
    repr(res)
    string = pickle.dumps(res, pickle.HIGHEST_PROTOCOL)
    resr = pickle.loads(string)
    tf_clusters = resr.find_clusters(pmin=0.05)
    peaks = resr.find_peaks()
    assert_dataobj_equal(tf_clusters, res.find_clusters(pmin=0.05))
    assert_dataobj_equal(peaks, res.find_peaks())
    assert tf_clusters.eval("p.min()") == peaks.eval("p.min()")
    unmasked = resr.f[0]
    masked = resr.masked_parameter_map(effect=0, pmin=0.05)
    assert_array_equal(masked.x <= unmasked.x, True)

    # reproducibility
    decimal = 12 if IS_WINDOWS else None  # FIXME: why is Windows sometimes different???
    res0 = testnd.ANOVA('utsnd', 'A*B*rm', data=ds, pmin=0.05, samples=5)
    res = testnd.ANOVA('utsnd', 'A*B*rm', data=ds, pmin=0.05, samples=5)
    assert_dataset_equal(res.clusters, res0.clusters, decimal=decimal)
    configure(n_workers=0)
    res = testnd.ANOVA('utsnd', 'A*B*rm', data=ds, pmin=0.05, samples=5)
    assert_dataset_equal(res.clusters, res0.clusters, decimal=decimal)
    configure(n_workers=True)

    # permutation
    eelbrain._stats.permutation._YIELD_ORIGINAL = 1
    samples = 4
    # raw
    res = testnd.ANOVA('utsnd', 'A*B*rm', data=ds, samples=samples)
    for dist in res._cdist:
        assert len(dist.dist) == samples
        assert_array_equal(dist.dist, dist.parameter_map.abs().max())
    # TFCE
    res = testnd.ANOVA('utsnd', 'A*B*rm', data=ds, tfce=True, samples=samples)
    for dist in res._cdist:
        assert len(dist.dist) == samples
        assert_array_equal(dist.dist, dist.tfce_map.abs().max())
    # thresholded
    res1 = testnd.ANOVA('utsnd', 'A*B*rm', data=ds, pmin=0.05, samples=samples)
    clusters = res1.find_clusters()
    for dist, effect in zip(res1._cdist, res1.effects):
        effect_idx = clusters.eval("effect == %r" % effect)
        vmax = clusters[effect_idx, 'v'].abs().max()
        assert len(dist.dist) == samples
        assert_array_equal(dist.dist, vmax)
    eelbrain._stats.permutation._YIELD_ORIGINAL = 0

    # 1d TFCE
    configure(n_workers=0)
    res = testnd.ANOVA('utsnd.rms(time=(0.1, 0.3))', 'A*B*rm', data=ds, tfce=True, samples=samples)
    configure(n_workers=True)

    # zero variance
    res2 = testnd.ANOVA('utsnd', 'A', data=ds)
    ds['utsnd'].x[:, 1, 10] = 0.
    zero_var = ds['utsnd'].var('case') == 0
    zv_index = tuple(i[0] for i in zero_var.nonzero())
    res1_zv = testnd.ANOVA('utsnd', 'A*B*rm', data=ds)
    res2_zv = testnd.ANOVA('utsnd', 'A', data=ds)
    for res, res_zv in ((res1, res1_zv), (res2, res2_zv)):
        for f, f_zv in zip(res.f, res_zv.f):
            assert_array_equal((f_zv == 0).x, zero_var.x)
            assert f_zv[zv_index] == 0
            f_zv[zv_index] = f[zv_index]
            assert_dataobj_equal(f_zv, f, decimal=decimal)

    # nested random effect
    res = testnd.ANOVA('uts', 'A * B * nrm(A)', data=ds, samples=10, tstart=.4)
    assert res.match == 'nrm(A)'
    assert [p.min() for p in res.p] == [0.0, 0.6, 0.9]

    # unequal argument length
    with pytest.raises(ValueError):
        testnd.ANOVA('uts', 'A[:-1]', data=ds)
    with pytest.raises(ValueError):
        testnd.ANOVA('uts[:-1]', 'A * B * nrm(A)', data=ds)


def test_anova_incremental():
    "Test testnd.ANOVA() with incremental f-tests"
    ds = datasets.get_uts()
    testnd.ANOVA('uts', 'A*B', data=ds[3:], pmin=0.05, samples=10)


@requires_mne_sample_data
def test_anova_parc():
    "Test ANOVA with parc argument and source space data"
    set_log_level('warning', 'mne')
    ds = datasets.get_mne_sample(src='ico', sub="side.isin(('L', 'R'))")
    y = ds['src'].sub(source=('lateraloccipital-lh', 'cuneus-lh'))
    y1 = y.sub(source='lateraloccipital-lh')
    y2 = y.sub(source='cuneus-lh')
    kwa = dict(ds=ds, tstart=0.2, tstop=0.3, samples=100)

    resp = testnd.ANOVA(y, "side*modality", pmin=0.05, parc='source', **kwa)
    c1p = resp.find_clusters(source='lateraloccipital-lh')
    c2p = resp.find_clusters(source='cuneus-lh')
    del c1p['p_parc', 'id']
    del c2p['p_parc', 'id']
    res1 = testnd.ANOVA(y1, "side*modality", pmin=0.05, **kwa)
    c1 = res1.find_clusters()
    del c1['id']
    res2 = testnd.ANOVA(y2, "side*modality", pmin=0.05, **kwa)
    c2 = res2.find_clusters()
    del c2['id']
    assert_dataset_equal(c1p, c1)
    assert_dataset_equal(c2p, c2)
    assert_array_equal(c2['p'], [0.85, 0.88, 0.97, 0.75, 0.99, 0.99, 0.98, 0.0,
                                 0.12, 0.88, 0.25, 0.97, 0.34, 0.96])

    # without multiprocessing
    configure(n_workers=0)
    ress = testnd.ANOVA(y, "side*modality", pmin=0.05, parc='source', **kwa)
    c1s = ress.find_clusters(source='lateraloccipital-lh')
    c2s = ress.find_clusters(source='cuneus-lh')
    del c1s['p_parc', 'id']
    del c2s['p_parc', 'id']
    assert_dataset_equal(c1s, c1)
    assert_dataset_equal(c2s, c2)
    configure(n_workers=True)

    # parc but single label
    resp2 = testnd.ANOVA(y2, "side*modality", pmin=0.05, parc='source', **kwa)
    c2sp = resp2.find_clusters(source='cuneus-lh')
    del c2sp['p_parc', 'id']
    assert_dataset_equal(c2sp, c2)

    # not defined
    with pytest.raises(NotImplementedError):
        testnd.ANOVA(y, "side*modality", tfce=True, parc='source', **kwa)


def test_clusterdist():
    "Test NDPermutationDistribution class"
    shape = (10, 6, 6, 4)
    locs = [[0, 0, 0],
            [1, 0, 0],
            [1, 1, 0],
            [0, 1, 0]]
    x = np.random.normal(0, 1, shape)
    sensor = Sensor(locs, ['0', '1', '2', '3'])
    sensor.set_adjacency(connect_dist=1.1)
    dims = ('case', UTS(-0.1, 0.1, 6), Scalar('dim2', range(6), 'unit'),
            sensor)
    y = NDVar(x, dims)

    # test connecting sensors
    logging.info("TEST:  connecting sensors")
    bin_map = np.zeros(shape[1:], bool)
    bin_map[:3, :3, :2] = True
    pmap = np.random.normal(0, 1, shape[1:])
    np.clip(pmap, -1, 1, pmap)
    pmap[bin_map] = 2
    cdist = NDPermutationDistribution(y, 0, 1.5)
    print(repr(cdist))
    cdist.add_original(pmap)
    print(repr(cdist))
    assert cdist.n_clusters == 1
    assert_array_equal(cdist._original_cluster_map == cdist._cids[0],
                       cdist._crop(bin_map).swapaxes(0, cdist._nad_ax))
    assert cdist.parameter_map.dims == y.dims[1:]

    # test connecting many sensors
    logging.info("TEST:  connecting sensors")
    bin_map = np.zeros(shape[1:], bool)
    bin_map[:3, :3] = True
    pmap = np.random.normal(0, 1, shape[1:])
    np.clip(pmap, -1, 1, pmap)
    pmap[bin_map] = 2
    cdist = NDPermutationDistribution(y, 0, 1.5)
    cdist.add_original(pmap)
    assert cdist.n_clusters == 1
    assert_array_equal(cdist._original_cluster_map == cdist._cids[0],
                       cdist._crop(bin_map).swapaxes(0, cdist._nad_ax))

    # test keeping sensors separate
    logging.info("TEST:  keeping sensors separate")
    bin_map = np.zeros(shape[1:], bool)
    bin_map[:3, :3, 0] = True
    bin_map[:3, :3, 2] = True
    pmap = np.random.normal(0, 1, shape[1:])
    np.clip(pmap, -1, 1, pmap)
    pmap[bin_map] = 2
    cdist = NDPermutationDistribution(y, 1, 1.5)
    cdist.add_original(pmap)
    assert cdist.n_clusters == 2

    # criteria
    ds = datasets.get_uts(True)
    res = testnd.TTestRelated('utsnd', 'A', match='rm', data=ds, samples=0, pmin=0.05)
    assert res.clusters['duration'].min() < 0.01
    assert res.clusters['n_sensors'].min() == 1
    res = testnd.TTestRelated('utsnd', 'A', match='rm', data=ds, samples=0, pmin=0.05,
                              mintime=0.02, minsensor=2)
    assert res.clusters['duration'].min() >= 0.02
    assert res.clusters['n_sensors'].min() == 2

    # 1d
    res1d = testnd.TTestRelated('utsnd.sub(time=0.1)', 'A', match='rm', data=ds,
                                samples=0, pmin=0.05)
    assert_dataobj_equal(res1d.p_uncorrected, res.p_uncorrected.sub(time=0.1))

    # TFCE
    logging.info("TEST:  TFCE")
    sensor = Sensor(locs, ['0', '1', '2', '3'])
    sensor.set_adjacency(connect_dist=1.1)
    time = UTS(-0.1, 0.1, 4)
    scalar = Scalar('scalar', range(10), 'unit')
    dims = ('case', time, sensor, scalar)
    rng = np.random.RandomState(0)
    y = NDVar(rng.normal(0, 1, (10, 4, 4, 10)), dims)
    cdist = NDPermutationDistribution(y, 3, None)
    cdist.add_original(y.x[0])
    cdist.finalize()
    assert cdist.dist.shape == (3,)
    # I/O
    string = pickle.dumps(cdist, pickle.HIGHEST_PROTOCOL)
    cdist_ = pickle.loads(string)
    assert repr(cdist_) == repr(cdist)
    # find peaks
    x = np.array([[[0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
                   [7, 7, 0, 0, 0, 0, 0, 0, 0, 0],
                   [0, 7, 0, 0, 0, 0, 0, 0, 0, 0],
                   [0, 0, 0, 0, 0, 0, 0, 0, 0, 0]],

                  [[0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
                   [5, 7, 0, 0, 0, 0, 0, 0, 0, 0],
                   [0, 6, 0, 0, 0, 0, 0, 0, 0, 0],
                   [0, 0, 0, 0, 0, 0, 0, 0, 0, 0]],

                  [[0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
                   [0, 0, 0, 0, 0, 7, 5, 5, 0, 0],
                   [0, 0, 0, 0, 5, 4, 4, 4, 0, 0],
                   [0, 0, 0, 0, 0, 0, 0, 0, 0, 0]],

                  [[0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
                   [0, 0, 0, 0, 0, 0, 0, 4, 0, 0],
                   [0, 0, 0, 0, 7, 0, 0, 3, 0, 0],
                   [0, 0, 0, 0, 0, 0, 0, 0, 0, 0]]])
    tgt = np.equal(x, 7)
    peaks = find_peaks(x, cdist._adjacency)
    logging.debug(' detected: \n%s' % (peaks.astype(int)))
    logging.debug(' target: \n%s' % (tgt.astype(int)))
    assert_array_equal(peaks, tgt)
    # testnd permutation result
    res = testnd.TTestOneSample(y, tfce=True, samples=3)
    if sys.version_info[0] == 3:
        target = [96.84232967, 205.83207424, 425.65942084]
    else:
        target = [77.5852307, 119.1976153, 217.6270428]
    assert_allclose(np.sort(res._cdist.dist), target)

    # parc with TFCE on unconnected dimension
    configure(False)
    x = rng.normal(0, 1, (10, 5, 2, 4))
    time = UTS(-0.1, 0.1, 5)
    categorial = Categorial('categorial', ('a', 'b'))
    y = NDVar(x, ('case', time, categorial, sensor))
    y0 = NDVar(x[:, :, 0], ('case', time, sensor))
    y1 = NDVar(x[:, :, 1], ('case', time, sensor))
    res = testnd.TTestOneSample(y, tfce=True, samples=3)
    res_parc = testnd.TTestOneSample(y, tfce=True, samples=3, parc='categorial')
    res0 = testnd.TTestOneSample(y0, tfce=True, samples=3)
    res1 = testnd.TTestOneSample(y1, tfce=True, samples=3)
    # cdist
    assert res._cdist.shape == (4, 2, 5)
    # T-maps don't depend on adjacency
    assert_array_equal(res.t.x[:, 0], res0.t.x)
    assert_array_equal(res.t.x[:, 1], res1.t.x)
    assert_array_equal(res_parc.t.x[:, 0], res0.t.x)
    assert_array_equal(res_parc.t.x[:, 1], res1.t.x)
    # TFCE-maps should always be the same because they're unconnected
    assert_array_equal(res.tfce_map.x[:, 0], res0.tfce_map.x)
    assert_array_equal(res.tfce_map.x[:, 1], res1.tfce_map.x)
    assert_array_equal(res_parc.tfce_map.x[:, 0], res0.tfce_map.x)
    assert_array_equal(res_parc.tfce_map.x[:, 1], res1.tfce_map.x)
    # Probability-maps should depend on what is taken into account
    p_a = res0.compute_probability_map().x
    p_b = res1.compute_probability_map().x
    assert_array_equal(res_parc.compute_probability_map(categorial='a').x, p_a)
    assert_array_equal(res_parc.compute_probability_map(categorial='b').x, p_b)
    p_parc = res_parc.compute_probability_map()
    assert_array_equal(p_parc.x, res.compute_probability_map().x)
    assert np.all(p_parc.sub(categorial='a').x >= p_a)
    assert np.all(p_parc.sub(categorial='b').x >= p_b)
    configure(True)


def test_corr():
    "Test testnd.Correlation()"
    ds = datasets.get_uts(True)

    # add correlation
    ds['utsnd'].x[:, 3:5, 50:65] += ds['Y'].x[:, None, None]

    res = testnd.Correlation('utsnd', 'Y', data=ds, samples=0)
    assert repr(res) == "<Correlation 'utsnd', 'Y', samples=0>"
    for s, t in product('01234', (0.1, 0.2, 0.35)):
        target = test.Correlation(ds['utsnd'].sub(sensor=s, time=t), ds['Y']).r
        assert res.r.sub(sensor=s, time=t) == pytest.approx(target)
    res = testnd.Correlation('utsnd', 'Y', 'rm', data=ds, samples=0)
    repr(res)
    res = testnd.Correlation('utsnd', 'Y', data=ds, samples=10, pmin=0.05)
    repr(res)
    res = testnd.Correlation('utsnd', 'Y', data=ds, samples=10, tfce=True)
    repr(res)

    # persistence
    string = pickle.dumps(res, protocol=pickle.HIGHEST_PROTOCOL)
    res_ = pickle.loads(string)
    assert repr(res_) == repr(res)
    assert_dataobj_equal(res.p_uncorrected, res_.p_uncorrected)
    assert_dataobj_equal(res.p, res_.p)


def test_t_contrast():
    ds = datasets.get_uts()

    # simple contrast
    res = testnd.TContrastRelated('uts', 'A', 'a1>a0', 'rm', data=ds, samples=10, pmin=0.05)
    assert repr(res) == "<TContrastRelated 'uts', 'A', 'a1>a0', match='rm', samples=10, pmin=0.05, 7 clusters, p < .001>"
    res_ = testnd.TTestRelated('uts', 'A', 'a1', 'a0', 'rm', data=ds)
    assert_array_equal(res.t.x, res_.t.x)

    # complex contrast
    res = testnd.TContrastRelated('uts', 'A%B', 'min(a0|b0>a1|b0, a0|b1>a1|b1)', 'rm', data=ds, samples=10, pmin=0.05)
    res_b0 = testnd.TTestRelated('uts', 'A%B', ('a0', 'b0'), ('a1', 'b0'), 'rm', data=ds)
    res_b1 = testnd.TTestRelated('uts', 'A%B', ('a0', 'b1'), ('a1', 'b1'), 'rm', data=ds)
    assert_array_equal(res.t.x, np.min([res_b0.t.x, res_b1.t.x], axis=0))

    # persistence
    string = pickle.dumps(res, protocol=pickle.HIGHEST_PROTOCOL)
    res_ = pickle.loads(string)
    assert repr(res_) == repr(res)
    assert_dataobj_equal(res.p, res_.p)

    # contrast with "*"
    res = testnd.TContrastRelated('uts', 'A%B', 'min(a1|b0>a0|b0, a1|b1>a0|b1)', 'rm', data=ds, tail=1, samples=0)

    # zero variance
    ds['uts'].x[:, 10] = 0.
    with pytest.raises(ZeroVarianceError):
        testnd.TContrastRelated('uts', 'A%B', 'min(a1|b0>a0|b0, a1|b1>a0|b1)', 'rm', tail=1, data=ds, samples=0)


def test_labeling():
    "Test cluster labeling"
    shape = (4, 20)
    pmap = np.empty(shape, float)
    edges = np.array([(0, 1), (0, 3), (1, 2), (2, 3)], np.uint32)
    conn = Adjacency((
        Scalar('graph', range(4), adjacency=edges),
        UTS(0, 0.01, 20)))
    criteria = None

    # some clusters
    pmap[:] = [[3, 3, 0, 0, 0, 0, 8, 0, 0, 0, 0, 0, 0, 0, 4, 4, 0, 0, 0, 0],
               [0, 1, 0, 0, 0, 0, 8, 0, 0, 4, 4, 4, 0, 0, 0, 0, 0, 0, 4, 0],
               [0, 3, 3, 4, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 4, 4],
               [0, 0, 0, 0, 0, 0, 8, 0, 0, 0, 0, 0, 4, 4, 4, 0, 0, 0, 0, 0]]
    cmap, cids = label_clusters(pmap, 2, 0, conn, criteria)
    assert len(cids) == 6
    assert_array_equal(cmap > 0, np.abs(pmap) > 2)

    # some other clusters
    pmap[:] = [[4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 0, 0],
               [0, 4, 0, 0, 0, 0, 0, 4, 0, 4, 4, 4, 0, 0, 0, 0, 0, 0, 0, 0],
               [0, 0, 4, 4, 0, 4, 4, 0, 4, 0, 0, 0, 4, 4, 1, 0, 4, 4, 0, 0],
               [0, 0, 0, 0, 4, 0, 0, 0, 0, 0, 0, 0, 0, 0, 4, 4, 0, 0, 0, 0]]
    cmap, cids = label_clusters(pmap, 2, 0, conn, criteria)
    assert len(cids) == 6
    assert_array_equal(cmap > 0, np.abs(pmap) > 2)


def test_ttest_1samp():
    "Test testnd.TTestOneSample()"
    ds = datasets.get_uts(True)

    # no clusters
    res0 = testnd.TTestOneSample('uts', sub="A == 'a0'", data=ds, samples=0)
    assert res0.p_uncorrected.min() < 0.05
    assert repr(res0) == "<TTestOneSample 'uts', sub=\"A == 'a0'\", samples=0>"

    # sub as array
    res1 = testnd.TTestOneSample('uts', sub=ds.eval("A == 'a0'"), data=ds, samples=0)
    assert repr(res1) == "<TTestOneSample 'uts', sub=<array>, samples=0>"

    # clusters without resampling
    res1 = testnd.TTestOneSample('uts', sub="A == 'a0'", data=ds, samples=0, pmin=0.05, tstart=0, tstop=0.6, mintime=0.05)
    assert res1.clusters.n_cases == 1
    assert 'p' not in res1.clusters
    assert repr(res1) == "<TTestOneSample 'uts', sub=\"A == 'a0'\", samples=0, pmin=0.05, tstart=0, tstop=0.6, mintime=0.05, 1 clusters>"

    # persistence
    string = pickle.dumps(res1, pickle.HIGHEST_PROTOCOL)
    res1_ = pickle.loads(string)
    assert repr(res1_) == repr(res1)
    assert_dataobj_equal(res1.p_uncorrected, res1_.p_uncorrected)

    # clusters with resampling
    res2 = testnd.TTestOneSample('uts', sub="A == 'a0'", data=ds, samples=10, pmin=0.05, tstart=0, tstop=0.6, mintime=0.05)
    assert res2.clusters.n_cases == 1
    assert res2.samples == 10
    assert 'p' in res2.clusters
    assert repr(res2) == "<TTestOneSample 'uts', sub=\"A == 'a0'\", samples=10, pmin=0.05, tstart=0, tstop=0.6, mintime=0.05, 1 clusters, p < .001>"

    # clusters with permutations
    dss = ds.sub("numpy.logical_and(A=='a0', B=='b0')")[:8]
    res3 = testnd.TTestOneSample('uts', sub="A == 'a0'", data=dss, samples=10000, pmin=0.05, tstart=0, tstop=0.6, mintime=0.05)
    assert repr(res3) == "<TTestOneSample 'uts', sub=\"A == 'a0'\", samples=255, pmin=0.05, tstart=0, tstop=0.6, mintime=0.05, 2 clusters, p = .020>"
    assert res3.clusters.n_cases == 2
    assert res3.samples == -1
    assert str(res3.clusters) == (
        '#   id   tstart   tstop   duration   v        p          sig\n'
        '------------------------------------------------------------\n'
        '0   3    0.08     0.34    0.26       95.692   0.019608   *  \n'
        '1   4    0.35     0.56    0.21       81.819   0.019608   *  \n'
        '------------------------------------------------------------\n'
        'NDVars: cluster')

    # nd
    dss = ds.sub("A == 'a0'")
    res = testnd.TTestOneSample('utsnd', data=dss, samples=1)
    res = testnd.TTestOneSample('utsnd', data=dss, pmin=0.05, samples=1)
    res = testnd.TTestOneSample('utsnd', data=dss, tfce=True, samples=1)

    # TFCE properties
    res = testnd.TTestOneSample('utsnd', sub="A == 'a0'", data=ds, samples=1)
    string = pickle.dumps(res, pickle.HIGHEST_PROTOCOL)
    res = pickle.loads(string)
    tfce_clusters = res.find_clusters(pmin=0.05)
    peaks = res.find_peaks()
    assert tfce_clusters.eval("p.min()") == peaks.eval("p.min()")
    masked = res.masked_parameter_map(pmin=0.05)
    assert_array_equal(masked.abs().x <= res.t.abs().x, True)

    # zero variance
    ds['utsnd'].x[:, 1, 10] = 0.
    ds['utsnd'].x[:, 2, 10] = 0.1
    res = testnd.TTestOneSample('utsnd', data=ds, samples=0)
    assert res.t.x[1, 10] == 0.
    assert res.t.x[2, 10] > 1e10

    # argument length
    with pytest.raises(ValueError):
        testnd.TTestOneSample('utsnd', sub="A[:-1] == 'a0'", data=ds, samples=0)


def test_ttest_ind():
    "Test testnd.TTestIndependent()"
    ds = datasets.get_uts(True)

    # basic
    res = testnd.TTestIndependent('uts', 'A', 'a1', 'a0', data=ds, samples=0)
    assert repr(res) == "<TTestIndependent 'uts', 'A', 'a1' (n=30), 'a0' (n=30), samples=0>"
    assert res.p_uncorrected.min() < 0.05
    # persistence
    string = pickle.dumps(res, pickle.HIGHEST_PROTOCOL)
    res_ = pickle.loads(string)
    assert repr(res_) == "<TTestIndependent 'uts', 'A', 'a1' (n=30), 'a0' (n=30), samples=0>"
    assert_dataobj_equal(res.p_uncorrected, res_.p_uncorrected)
    # alternate argspec
    res_ = testnd.TTestIndependent("uts[A == 'a1']", "uts[A == 'a0']", data=ds, samples=0)
    assert repr(res_) == "<TTestIndependent 'uts' (n=30), 'uts' (n=30), samples=0>"
    assert_dataobj_equal(res_.t, res.t)

    # cluster
    res = testnd.TTestIndependent('uts', 'A', 'a1', 'a0', data=ds, tail=1, samples=1)
    # persistence
    string = pickle.dumps(res, pickle.HIGHEST_PROTOCOL)
    res_ = pickle.loads(string)
    assert repr(res_) == repr(res)
    assert_dataobj_equal(res.p_uncorrected, res_.p_uncorrected)

    # nd
    res = testnd.TTestIndependent('utsnd', 'A', 'a1', 'a0', data=ds, pmin=0.05, samples=2)
    assert res._cdist.n_clusters == 10

    # zero variance
    ds['utsnd'].x[:, 1, 10] = 0.
    res_zv = testnd.TTestIndependent('utsnd', 'A', 'a1', 'a0', data=ds, samples=0)
    assert_array_equal(res_zv.t.x[0], res.t.x[0])
    assert res_zv.t.x[1, 10] == 0.
    # argument mismatch
    with pytest.raises(ValueError):
        testnd.TTestIndependent(ds['utsnd'], ds[:-1, 'A'], samples=0)

    # large array (#52)
    ds = datasets.simulate_erp(60, time=UTS(-0.1, 0.001, 601), snr=0.7)
    res = testnd.TTestIndependent('eeg', 'predictability', 'low', 'high', data=ds, pmin=0.05, tstart=0.100, mintime=0.020, samples=20)
    clusters = res.find_clusters()
    assert_array_equal(clusters['p'], [0.4, 0.2, 0.4, 0.2, 0. ])


def test_ttest_rel():
    "Test testnd.TTestRelated()"
    ds = datasets.get_uts(True)

    # basic
    res = testnd.TTestRelated('uts', 'A%B', ('a1', 'b1'), ('a0', 'b0'), 'rm', data=ds, samples=100)
    assert repr(res) == "<TTestRelated 'uts', 'A x B', ('a1', 'b1'), ('a0', 'b0'), 'rm' (n=15), samples=100, p < .001>"
    difference = res.masked_difference()
    assert difference.x.mask.sum() == 84
    c1 = res.masked_c1()
    assert c1.x.mask.sum() == 84
    assert_array_equal(c1.x.data, res.c1_mean.x)

    # alternate argspec
    res_ = testnd.TTestRelated("uts[A%B == ('a1', 'b1')]", "uts[A%B == ('a0', 'b0')]", data=ds, samples=100)
    assert repr(res_) == "<TTestRelated 'uts', 'uts' (n=15), samples=100, p < .001>"
    assert_dataobj_equal(res_.t, res.t)
    # alternate argspec 2
    ds1 = Dataset()
    ds1['a1b1'] = ds.eval("uts[A%B == ('a1', 'b1')]")
    ds1['a0b0'] = ds.eval("uts[A%B == ('a0', 'b0')]")
    res1 = testnd.TTestRelated('a1b1', 'a0b0', data=ds1, samples=100)
    assert_dataobj_equal(res1.t, res.t)
    assert repr(res1) == "<TTestRelated 'a1b1', 'a0b0' (n=15), samples=100, p < .001>"

    # persistence
    string = pickle.dumps(res, pickle.HIGHEST_PROTOCOL)
    res_ = pickle.loads(string)
    assert repr(res_) == repr(res)
    assert_dataobj_equal(res.p_uncorrected, res_.p_uncorrected)

    # collapsing cells
    res2 = testnd.TTestRelated('uts', 'A', 'a1', 'a0', 'rm', data=ds, samples=0)
    assert res2.p_uncorrected.min() < 0.05
    assert res2.n == res.n

    # reproducibility
    res3 = testnd.TTestRelated('uts', 'A%B', ('a1', 'b1'), ('a0', 'b0'), 'rm', data=ds, samples=100)
    assert_dataset_equal(res3.find_clusters(maps=True), res.clusters)
    configure(n_workers=0)
    res4 = testnd.TTestRelated('uts', 'A%B', ('a1', 'b1'), ('a0', 'b0'), 'rm', data=ds, samples=100)
    assert_dataset_equal(res4.find_clusters(maps=True), res.clusters)
    configure(n_workers=True)
    sds = ds.sub("B=='b0'")
    # thresholded, UTS
    configure(n_workers=0)
    res0 = testnd.TTestRelated('uts', 'A', 'a1', 'a0', 'rm', data=sds, pmin=0.1, samples=100)
    tgt = res0.find_clusters()
    configure(n_workers=True)
    res1 = testnd.TTestRelated('uts', 'A', 'a1', 'a0', 'rm', data=sds, pmin=0.1, samples=100)
    assert_dataset_equal(res1.find_clusters(), tgt)
    # thresholded, UTSND
    configure(n_workers=0)
    res0 = testnd.TTestRelated('utsnd', 'A', 'a1', 'a0', 'rm', data=sds, pmin=0.1, samples=100)
    tgt = res0.find_clusters()
    configure(n_workers=True)
    res1 = testnd.TTestRelated('utsnd', 'A', 'a1', 'a0', 'rm', data=sds, pmin=0.1, samples=100)
    assert_dataset_equal(res1.find_clusters(), tgt)
    # TFCE, UTS
    configure(n_workers=0)
    res0 = testnd.TTestRelated('uts', 'A', 'a1', 'a0', 'rm', data=sds, tfce=True, samples=10)
    tgt = res0.compute_probability_map()
    configure(n_workers=True)
    res1 = testnd.TTestRelated('uts', 'A', 'a1', 'a0', 'rm', data=sds, tfce=True, samples=10)
    assert_dataobj_equal(res1.compute_probability_map(), tgt)

    # zero variance
    ds['utsnd'].x[:, 1, 10] = 0.
    res = testnd.TTestRelated('utsnd', 'A', match='rm', data=ds)
    assert res.t.x[1, 10] == 0

    # argument length
    with pytest.raises(ValueError):
        testnd.TTestRelated('utsnd', 'A[:-1]', match='rm', data=ds)
    with pytest.raises(ValueError):
        testnd.TTestRelated('utsnd', 'A', match='rm[:-1]', data=ds)


def test_vector():
    """Test vector tests"""
    # single vector
    ds = datasets.get_uv(vector=True)
    res = testnd.Vector('v[:40]', data=ds, samples=10)
    assert res.p == 0.0
    res = testnd.Vector('v[40:]', data=ds, samples=10)
    assert res.p == 1.0

    # single vector with norm stat
    res_t = testnd.Vector('v[:40]', data=ds, samples=10, norm=True)
    assert res_t.p == 0.0
    res_t = testnd.Vector('v[40:]', data=ds, samples=10, norm=True)
    assert res_t.p == 1.0

    # non-space tests should raise error
    with pytest.raises(WrongDimensionError):
        testnd.TTestOneSample('v', data=ds)
    with pytest.raises(WrongDimensionError):
        testnd.TTestRelated('v', 'A', match='rm', data=ds)
    with pytest.raises(WrongDimensionError):
        testnd.TTestIndependent('v', 'A', data=ds)
    with pytest.raises(WrongDimensionError):
        testnd.TContrastRelated('v', 'A', 'a0 > a1', 'rm', data=ds)
    with pytest.raises(WrongDimensionError):
        testnd.Correlation('v', 'fltvar', data=ds)
    with pytest.raises(WrongDimensionError):
        testnd.ANOVA('v', 'A * B', data=ds)

    # vector in time
    ds = datasets.get_uts(vector3d=True)
    v1 = ds[30:, 'v3d']
    v2 = ds[:30, 'v3d']
    vd = v1 - v2
    res = testnd.Vector(vd, samples=10)
    assert res.p.min() == 0
    difference = res.masked_difference(0.5)
    assert difference.x.mask.sum() == 288
    res_r = pickle.loads(pickle.dumps(res))
    assert repr(res_r) == repr(res)
    # diff related
    resd = testnd.VectorDifferenceRelated(v1, v2, samples=10)
    assert_dataobj_equal(resd.p, res.p, name=False)
    assert_dataobj_equal(resd.t2, res.t2, name=False)
    res_r = pickle.loads(pickle.dumps(resd))
    assert repr(res_r) == repr(resd)
    # diff independent
    res = VectorDifferenceIndependent(v1, v2, samples=10, norm=True)
    assert_dataobj_equal(res.difference, v1.mean('case') - v2.mean('case'), name=False)
    assert res.p.max() == 1
    assert res.p.min() == 0
    res_r = pickle.loads(pickle.dumps(res))
    assert repr(res_r) == repr(res)
    # with mp
    res = testnd.Vector(v1, samples=10)
    assert res.p.min() == 0.1
    # without mp
    configure(n_workers=0)
    res0 = testnd.Vector(v1, samples=10)
    assert_array_equal(np.sort(res0._cdist.dist), np.sort(res._cdist.dist))
    configure(n_workers=True)
    # time window
    res = testnd.Vector(v2, samples=10, tstart=0.1, tstop=0.4)
    assert res.p.min() == 0.2
    difference = res.masked_difference(0.5)
    assert difference.x.mask.sum() == 297

    # vector in time with norm stat
    res = testnd.Vector(vd, samples=10, norm=True)
    assert res.p.min() == 0
    difference = res.masked_difference()
    assert difference.x.mask.sum() == 285
    resd = testnd.VectorDifferenceRelated(v1, v2, samples=10, norm=True)
    assert_dataobj_equal(resd.p, res.p, name=False)
    assert_dataobj_equal(resd.difference, res.difference, name=False)

    v_small = v2 / 100
    res = testnd.Vector(v_small, tfce=True, samples=10, norm=True)
    assert 'WARNING' in repr(res)
    res = testnd.Vector(v_small, tfce=0.1, samples=10)
    assert res.p.min() == 0.0


def test_cwt():
    "Test tests with wavelet transform"
    ds = datasets.get_uts(True)
    ds['cwt'] = cwt_morlet(ds['utsnd'], np.arange(10, 20)).abs()
    res = testnd.TTestRelated('cwt', 'A', match='rm', data=ds, pmin=0.05, samples=10)
    cluster = res.clusters.sub("p == 0")
    assert_array_equal(cluster['frequency_min'], 10)
    assert_array_equal(cluster['frequency_max'], 19)


def test_merged_temporal_cluster_dist():
    "Test use of _MergedTemporalClusterDist with testnd test results"
    ds1 = datasets.get_uts()
    ds2 = datasets.get_uts(seed=42)

    anova_kw = dict(y='uts', x='A*B*rm', match='rm', pmin=0.05, samples=10)
    ttest_kw = dict(y='uts', x='A', c1='a1', c0='a0', pmin=0.05, samples=10)
    contrast_kw = dict(y='uts', x='A', contrast='a1>a0', pmin=0.05, samples=10)

    def test_merged(res1, res2):
        merged_dist = _MergedTemporalClusterDist([res1._cdist, res2._cdist])
        if isinstance(res1, testnd.ANOVA):
            assert len(merged_dist.dist) == len(res1.effects)
            for effect, dist in merged_dist.dist.items():
                assert effect in res1.effects
                assert len(dist) == res1.samples
        else:
            assert len(merged_dist.dist) == res1.samples
        res1_clusters = merged_dist.correct_cluster_p(res1)
        res2_clusters = merged_dist.correct_cluster_p(res2)
        for clusters in [res1_clusters, res2_clusters]:
            assert 'p_parc' in clusters
            for cl in clusters.itercases():
                assert cl['p_parc'] >= cl['p']

    # multi-effect
    res1 = testnd.ANOVA(data=ds1, **anova_kw)
    res2 = testnd.ANOVA(data=ds2, **anova_kw)
    test_merged(res1, res2)

    # TTestRelated
    res1 = testnd.TTestRelated(data=ds1, match='rm', **ttest_kw)
    res2 = testnd.TTestRelated(data=ds2, match='rm', **ttest_kw)
    test_merged(res1, res2)

    # TTestIndependent
    res1 = testnd.TTestIndependent(data=ds1, **ttest_kw)
    res2 = testnd.TTestIndependent(data=ds2, **ttest_kw)
    test_merged(res1, res2)

    # TTestOneSample
    res1 = testnd.TTestOneSample('uts', data=ds1, pmin=0.05, samples=10)
    res2 = testnd.TTestOneSample('uts', data=ds2, pmin=0.05, samples=10)
    test_merged(res1, res2)

    # TContrastRelated
    res1 = testnd.TContrastRelated(data=ds1, match='rm', **contrast_kw)
    res2 = testnd.TContrastRelated(data=ds2, match='rm', **contrast_kw)
    test_merged(res1, res2)



# Author: Christian Brodbeck <christianbrodbeck@nyu.edu>
"""Test MneExperiment using mne-python sample data"""
from os.path import join, exists
import pytest
from warnings import catch_warnings, filterwarnings

import numpy as np
from numpy.testing import assert_almost_equal, assert_array_equal

from eelbrain import *
from eelbrain.pipeline import *
from eelbrain._exceptions import DefinitionError
from eelbrain.testing import TempDir, assert_dataobj_equal, requires_mne_sample_data


@requires_mne_sample_data
def test_sample():
    set_log_level('warning', 'mne')
    from eelbrain._experiment.tests.sample_experiment import SampleExperiment

    tempdir = TempDir()
    datasets.setup_samples_experiment(tempdir, n_subjects=3, n_segments=2)

    root = join(tempdir, 'SampleExperiment')
    e = SampleExperiment(root)

    assert e.get('raw') == '1-40'
    assert e.get('subject') == 'R0000'
    assert e.get('subject', subject='R0002') == 'R0002'

    # events
    e.set('R0001', rej='')
    ds = e.load_selected_events(epoch='target')
    assert ds.n_cases == 39
    ds = e.load_selected_events(epoch='auditory')
    assert ds.n_cases == 20
    ds = e.load_selected_events(epoch='av')
    assert ds.n_cases == 39

    # evoked cache invalidated by change in bads
    e.set('R0001', rej='', epoch='target')
    ds = e.load_evoked(ndvar=False)
    assert ds[0, 'evoked'].info['bads'] == []
    e.make_bad_channels(['MEG 0331'])
    ds = e.load_evoked(ndvar=False)
    assert ds[0, 'evoked'].info['bads'] == ['MEG 0331']

    e.set(rej='man', model='modality')
    sds = []
    for _ in e:
        e.make_epoch_selection(auto=2.5e-12)
        sds.append(e.load_evoked())
    ds_ind = combine(sds, dim_intersection=True)

    ds = e.load_evoked('all')
    ds['meg'] = ds['meg'].sub(sensor=ds['meg'].sensor.index(exclude='MEG 0331'))  # load_evoked interpolates bad channel
    assert_dataobj_equal(ds_ind, ds, decimal=19)  # make vs load evoked

    # sensor space tests
    megs = [e.load_evoked(cat='auditory')['meg'] for _ in e]
    res = e.load_test('a>v', 0.05, 0.2, 0.05, samples=100, data='sensor.rms', baseline=False, make=True)
    meg_rms = combine(meg.rms('sensor') for meg in megs).mean('case', name='auditory')
    assert_dataobj_equal(res.c1_mean, meg_rms, decimal=21)
    res = e.load_test('a>v', 0.05, 0.2, 0.05, samples=100, data='sensor.mean', baseline=False, make=True)
    meg_mean = combine(meg.mean('sensor') for meg in megs).mean('case', name='auditory')
    assert_dataobj_equal(res.c1_mean, meg_mean, decimal=21)
    with pytest.raises(IOError):
        e.load_test('a>v', 0.05, 0.2, 0.05, samples=20, data='sensor', baseline=False)
    res = e.load_test('a>v', 0.05, 0.2, 0.05, samples=20, data='sensor', baseline=False, make=True)
    assert res.p.min() == pytest.approx(.143, abs=.001)
    assert res.difference.max() == pytest.approx(4.47e-13, 1e-15)
    # plot (skip to avoid using framework build)
    # e.plot_evoked(1, epoch='target', model='')

    # e._report_subject_info() broke with non-alphabetic subject order
    subjects = e.get_field_values('subject')
    ds = Dataset()
    ds['subject'] = Factor(reversed(subjects))
    ds['n'] = Var(range(3))
    _ = e._report_subject_info(ds, '')

    # post_baseline_trigger_shift
    # use multiple of tstep to shift by even number of samples
    tstep = 0.008324800548266162
    shift = -7 * tstep

    class Experiment(SampleExperiment):
        epochs = {
            **SampleExperiment.epochs,
            'visual-s': SecondaryEpoch('target', "modality == 'visual'", post_baseline_trigger_shift='shift', post_baseline_trigger_shift_max=0, post_baseline_trigger_shift_min=shift),
        }
        variables = {
            **SampleExperiment.variables,
            'shift': LabelVar('side', {'left': 0, 'right': shift}),
            'shift_t': LabelVar('trigger', {(1, 3): 0, (2, 4): shift})
        }
    e = Experiment(root)
    # test shift in events
    ds = e.load_events()
    assert_dataobj_equal(ds['shift_t'], ds['shift'], name=False)
    # compare against epochs (baseline correction on epoch level rather than evoked for smaller numerical error)
    ep = e.load_epochs(baseline=True, epoch='visual', rej='').aggregate('side')
    evs = e.load_evoked(baseline=True, epoch='visual-s', rej='', model='side')
    tstart = ep['meg'].time.tmin - shift
    assert_dataobj_equal(evs[0, 'meg'], ep[0, 'meg'].sub(time=(tstart, None)), decimal=19)
    tstop = ep['meg'].time.tstop + shift
    assert_almost_equal(evs[1, 'meg'].x, ep[1, 'meg'].sub(time=(None, tstop)).x, decimal=19)

    # post_baseline_trigger_shift
    class Experiment(SampleExperiment):
        epochs = {
            **SampleExperiment.epochs,
            'v_shift': SecondaryEpoch('visual', vars={'shift': 'Var([0.0], repeat=len(side))'}),
            'a_shift': SecondaryEpoch('auditory', vars={'shift': 'Var([0.1], repeat=len(side))'}),
            'av_shift': SuperEpoch(('v_shift', 'a_shift'), post_baseline_trigger_shift='shift', post_baseline_trigger_shift_max=0.1, post_baseline_trigger_shift_min=0.0),
        }
        groups = {
            'group0': Group(['R0000']),
            'group1': SubGroup('all', ['R0000']),
        }
        variables = {
            'group': GroupVar(['group0', 'group1']),
            **SampleExperiment.variables,
        }
    e = Experiment(root)
    events = e.load_selected_events(epoch='av_shift')
    ds = e.load_epochs(baseline=True, epoch='av_shift')
    v = ds.sub("epoch=='v_shift'", 'meg')
    v_target = e.load_epochs(baseline=True, epoch='visual')['meg'].sub(time=(-0.1, v.time.tstop))
    assert_almost_equal(v.x, v_target.x)
    a = ds.sub("epoch=='a_shift'", 'meg').sub(time=(-0.1, 0.099))
    a_target = e.load_epochs(baseline=True, epoch='auditory')['meg'].sub(time=(0, 0.199))
    assert_almost_equal(a.x, a_target.x, decimal=20)

    # duplicate subject
    class BadExperiment(SampleExperiment):
        groups = {'group': ('R0001', 'R0002', 'R0002')}
    with pytest.raises(DefinitionError):
        BadExperiment(root)

    # non-existing subject
    class BadExperiment(SampleExperiment):
        groups = {'group': ('R0001', 'R0003', 'R0002')}
    with pytest.raises(DefinitionError):
        BadExperiment(root)

    # unsorted subjects
    class Experiment(SampleExperiment):
        groups = {'group': ('R0002', 'R0000', 'R0001')}
    e = Experiment(root)
    assert [s for s in e] == ['R0000', 'R0001', 'R0002']

    # changes
    class Changed(SampleExperiment):
        variables = {
            'event': {(1, 2, 3, 4): 'target', 5: 'smiley', 32: 'button'},
            'side': {(1, 3): 'left', (2, 4): 'right_changed'},
            'modality': {(1, 2): 'auditory', (3, 4): 'visual'}
        }
        tests = {
            'twostage': TwoStageTest(
                'side_left + modality_a',
                {'side_left': "side == 'left'",
                 'modality_a': "modality == 'auditory'"}),
            'novars': TwoStageTest('side + modality'),
        }
    e = Changed(root)

    # changed variable, while a test with model=None is not changed
    class Changed(Changed):
        variables = {
            'side': {(1, 3): 'left', (2, 4): 'right_changed'},
            'modality': {(1, 2): 'auditory', (3, 4): 'visual_changed'}
        }
    e = Changed(root)

    # changed variable, unchanged test with vardef=None
    class Changed(Changed):
        variables = {
            'side': {(1, 3): 'left', (2, 4): 'right_changed'},
            'modality': {(1, 2): 'auditory', (3, 4): 'visual_changed'}
        }
    e = Changed(root)

    # ICA
    # ---
    class Experiment(SampleExperiment):
        raw = {
            'apply-ica': RawApplyICA('tsss', 'ica'),
            **SampleExperiment.raw,
        }
    e = Experiment(root)
    ica_path = e.make_ica(raw='ica')
    e.set(raw='ica1-40', model='')
    e.make_epoch_selection(auto=2e-12, overwrite=True)
    ds1 = e.load_evoked(raw='ica1-40')
    ica = e.load_ica(raw='ica')
    ica.exclude = [0, 1, 2]
    ica.save(ica_path, overwrite=True)
    ds2 = e.load_evoked(raw='ica1-40')
    assert not np.allclose(ds1['meg'].x, ds2['meg'].x, atol=1e-20), "ICA change ignored"
    # apply-ICA
    with catch_warnings():
        filterwarnings('ignore', "The measurement information indicates a low-pass frequency", RuntimeWarning)
        ds1 = e.load_evoked(raw='ica', rej='')
        ds2 = e.load_evoked(raw='apply-ica', rej='')
    assert_dataobj_equal(ds2, ds1)

    # rename subject
    # --------------
    # e.set(subject='R0001')
    # src = Path(e._bids_path.directory)
    # dst = Path(str(src).replace('R0001', 'R0003'))
    # shutil.move(src, dst)
    # for path in dst.glob('*.fif'):
    #     shutil.move(path, dst / path.parent / path.name.replace('R0001', 'R0003'))
    # check subject list
    # e = SampleExperiment(root)
    # assert list(e) == ['R0000', 'R0002', 'R0003']
    # check that cached test got deleted
    # assert e.get('raw') == '1-40'
    # with pytest.raises(IOError):
    #     e.load_test('a>v', 0.05, 0.2, 0.05, samples=20, data='sensor', baseline=False)
    # res = e.load_test('a>v', 0.05, 0.2, 0.05, samples=20, data='sensor', baseline=False, make=True)
    # assert res.df == 2
    # assert res.p.min() == pytest.approx(.143, abs=.001)
    # assert res.difference.max() == pytest.approx(4.47e-13, 1e-15)

    # remove subject
    # --------------
    # shutil.rmtree(dst)
    # # check cache
    # e = SampleExperiment(root)
    # assert list(e) == ['R0000', 'R0002']
    # # check that cached test got deleted
    # assert e.get('raw') == '1-40'
    # with pytest.raises(IOError):
    #     e.load_test('a>v', 0.05, 0.2, 0.05, samples=20, data='sensor', baseline=False)

    # label_events
    # ------------
    class Experiment(SampleExperiment):
        def label_events(self, ds):
            SampleExperiment.label_events(self, ds)
            ds = ds.sub("event == 'smiley'")
            ds['new_var'] = Var([i + 1 for i in ds['i_start']])
            return ds

    e = Experiment(root)
    events = e.load_events()
    assert_array_equal(events['new_var'], [67402, 75306])

    # Parc
    # ----
    labels = e.load_annot(parc='ac', mrisubject='fsaverage')
    assert len(labels) == 4
    # change parc definition

    class Experiment(SampleExperiment):
        parcs = {
            'ac': SubParc('aparc', ('transversetemporal', 'superiortemporal')),
        }
    e = Experiment(root)
    assert len(e.glob('annot-file', True, parc='ac')) == 0
    labels = e.load_annot(parc='ac', mrisubject='fsaverage')
    assert len(labels) == 6


@requires_mne_sample_data
@pytest.mark.slow
def test_sample_source():
    set_log_level('warning', 'mne')
    from eelbrain._experiment.tests.sample_experiment import SampleExperiment

    tempdir = TempDir()
    datasets.setup_samples_experiment(tempdir, 3, 2, mris=True)  # TODO: use sample MRI which already has forward solution
    root = join(tempdir, 'SampleExperiment')
    e = SampleExperiment(root)

    # source space tests
    e.set(src='ico-4', rej='', epoch='auditory')
    # These two tests are only identical if the evoked has been cached before the first test is loaded
    resp = e.load_test('left=right', 0.05, 0.2, 0.05, samples=100, parc='ac', make=True)
    resm = e.load_test('left=right', 0.05, 0.2, 0.05, samples=100, mask='ac', make=True)
    assert_dataobj_equal(resp.t, resm.t)
    # ROI tests
    e.set(epoch='target')
    ress = e.load_test('left=right', 0.05, 0.2, 0.05, samples=100, data='source.rms', parc='ac', make=True)
    res = ress.res['transversetemporal-lh']
    assert res.p.min() == pytest.approx(0.429, abs=.001)
    ress = e.load_test('twostage', 0.05, 0.2, 0.05, samples=100, data='source.rms', parc='ac', make=True)
    res = ress.res['transversetemporal-lh']
    assert res.samples == -1
    assert res.tests['intercept'].p.min() == 1 / 7


@requires_mne_sample_data
def test_sample_tasks():
    set_log_level('warning', 'mne')
    from eelbrain._experiment.tests.sample_experiment_sessions import SampleExperiment

    tempdir = TempDir()
    datasets.setup_samples_experiment(tempdir, 2, 2, 1)

    class Experiment(SampleExperiment):

        raw = {
            'ica': RawICA('raw', ('sample1', 'sample2'), 'fastica', max_iter=1),
            **SampleExperiment.raw,
        }

    root = join(tempdir, 'SampleExperiment')
    e = Experiment(root)
    # bad channels
    e.make_bad_channels('0111')
    assert e.load_bad_channels() == ['MEG 0111']
    assert e.load_bad_channels(task='sample2') == []
    e.show_bad_channels()
    e.merge_bad_channels()
    assert e.load_bad_channels(task='sample2') == ['MEG 0111']
    e.show_bad_channels()

    # rejection
    for _ in e:
        for epoch in ('target1', 'target2'):
            e.set(epoch=epoch)
            e.make_epoch_selection(auto=2e-12)

    ds = e.load_evoked('R0000', epoch='target2')
    e.set(task='sample1')
    ds2 = e.load_evoked('R0000')
    assert_dataobj_equal(ds2, ds, decimal=19)

    # super-epoch
    ds1 = e.load_epochs(epoch='target1')
    ds2 = e.load_epochs(epoch='target2')
    ds_super = e.load_epochs(epoch='super')
    assert_dataobj_equal(ds_super['meg'], combine((ds1['meg'], ds2['meg'])))
    # evoked
    dse_super = e.load_evoked(epoch='super', model='modality%side')
    target = ds_super.aggregate('modality%side', drop=('i_start', 't_edf', 'T', 'index', 'trigger', 'task', 'interpolate_channels', 'epoch'))
    assert_dataobj_equal(dse_super, target, 19)

    # conflicting task and epoch settings
    rej_path = join(root, 'derivatives', 'eelbrain', 'epoch selection', 'sub-R0000_meg_raw-1-40_epoch-target2_rej-man_epoch.pickle')
    e.set(epoch='target2', raw='1-40')
    assert not exists(rej_path)
    e.set(task='sample1')
    e.make_epoch_selection(auto=2e-12)
    assert exists(rej_path)

    # ica
    e.set('R0000', raw='ica')
    with catch_warnings():
        filterwarnings('ignore', "FastICA did not converge", UserWarning)
        assert e.make_ica() == join(root, 'derivatives', 'ica', 'sub-R0000_meg_raw-ica_ica.fif')


@requires_mne_sample_data
def test_sample_neuromag():
    set_log_level('warning', 'mne')
    from eelbrain._experiment.tests.sample_experiment import SampleExperiment

    tempdir = TempDir()
    datasets.setup_samples_experiment(tempdir, n_subjects=1, pick='')

    class Experiment(SampleExperiment):
        defaults = {'raw': '1-40'}
        # raw = {'1-40': RawFilter('raw', 1, 40)}

    root = join(tempdir, 'SampleExperiment')
    e = Experiment(root)
    assert e.get('raw') == '1-40'

    e.make_epoch_selection(auto={'mag': 2e-12, 'grad': 5e-11, 'eeg': 1.5e-4})
    ds = e.load_selected_events(reject='keep')
    assert ds['accept'].sum() == 69
